/*
 Copyright (c) 2011 Technische Universitt Darmstadt.
 
 Permission is hereby granted to use, copy, modify, and distribute this
 software, provided that this copyright notice remains intact.
 */

package gsi.rf.data;

/**
 * This class represents an exception that is thrown
 * while trying to read a data file for analysis.
 * 
 * @author Christopher Spies, TU Darmstadt
 * @since 2011-05-23
 * @version 1.1.0.0
 */
public class ParseException extends Exception {

	////////////////////
	// private constants
	
	/**
	 * The Serial Version UID of this class, required by all serializable classes.
	 */
	private static final long serialVersionUID = -6887340033118520843L;
	
	////////////////////
	// constructors
	
	/**
	 * Create an exception object!
	 * 
	 * @param fileName The name of the offending file.
	 * @param message A detailed description of the exception conditions.
	 * 
	 * @since 2011-05-23
	 */
	public ParseException(String fileName, String message) {
		super(formatMessage(fileName, message));
	}
	
	/**
	 * Create an exception object!
	 * 
	 * @param fileName The name of the offending file.
	 * @param cause The underlying cause of the exception.
	 * 
	 * @since 2011-05-23
	 */
	public ParseException(String fileName, Throwable cause) {
		this(fileName, cause.getMessage(), cause);
	}
	
	/**
	 * Create an exception object!
	 * 
	 * @param fileName The name of the offending file.
	 * @param message A detailed description of the exception conditions.
	 * @param cause The underlying cause of the exception.
	 * 
	 * @since 2011-05-23
	 */
	public ParseException(String fileName, String message, Throwable cause) {
		super(formatMessage(fileName, message), cause);
	}
	
	/**
	 * Create an exception object!
	 * 
	 * @param fileName The name of the offending file.
	 * @param lineNumber The line number in the file at which the exception occurred. Line numbers start at 1.
	 * @param message A detailed description of the exception conditions.
	 * 
	 * @since 2011-05-23
	 */
	public ParseException(String fileName, int lineNumber, String message) {
		super(formatMessage(fileName, lineNumber, message));
	}
	
	/**
	 * Create an exception object!
	 * 
	 * @param fileName The name of the offending file.
	 * @param lineNumber The line number in the file at which the exception occurred. Line numbers start at 1.
	 * @param cause The underlying cause of the exception.
	 * 
	 * @since 2011-05-23
	 */
	public ParseException(String fileName, int lineNumber, Throwable cause) {
		this(fileName, lineNumber, cause.getMessage(), cause);
	}
	
	/**
	 * Create an exception object!
	 * 
	 * @param fileName The name of the offending file.
	 * @param lineNumber The line number in the file at which the exception occurred. Line numbers start at 1.
	 * @param message A detailed description of the exception conditions.
	 * @param cause The underlying cause of the exception.
	 * 
	 * @since 2011-05-23
	 */
	public ParseException(String fileName, int lineNumber, String message, Throwable cause) {
		super(formatMessage(fileName, lineNumber, message), cause);
	}
	
	/**
	 * Create an exception object!
	 * 
	 * @param documentName The name of the offending document.
	 * @param rowNumber The row number in the document at which the exception occurred. Row numbers start at 0, but are reported as starting from 1, i. e. the reported row number is the actual row number + 1.
	 * @param columnNumber The column number in the document at which the exception occurred. Column numbers start at 0, but are reported as starting from 1, i. e. the reported column number is the actual column number + 1.
	 * @param message A detailed description of the exception conditions.
	 * 
	 * @since 2011-06-09
	 */
	public ParseException(String documentName, int rowNumber, int columnNumber, String message) {
		super(formatMessage(documentName, rowNumber, columnNumber, message));
	}
	
	/**
	 * Create an exception object!
	 * 
	 * @param documentName The name of the offending document.
	 * @param rowNumber The row number in the document at which the exception occurred. Row numbers start at 0, but are reported as starting from 1, i. e. the reported row number is the actual row number + 1.
	 * @param columnNumber The column number in the document at which the exception occurred. Column numbers start at 0, but are reported as starting from 1, i. e. the reported column number is the actual column number + 1.
	 * @param cause The underlying cause of the exception.
	 * 
	 * @since 2011-06-09
	 */
	public ParseException(String documentName, int rowNumber, int columnNumber, Throwable cause) {
		this(documentName, rowNumber, columnNumber, cause.getMessage(), cause);
	}
	
	/**
	 * Create an exception object!
	 * 
	 * @param documentName The name of the offending document.
	 * @param rowNumber The row number in the document at which the exception occurred. Row numbers start at 0, but are reported as starting from 1, i. e. the reported row number is the actual row number + 1.
	 * @param columnNumber The column number in the document at which the exception occurred. Column numbers start at 0, but are reported as starting from 1, i. e. the reported column number is the actual column number + 1.
	 * @param message A detailed description of the exception conditions.
	 * @param cause The underlying cause of the exception.
	 * 
	 * @since 2011-06-09
	 */
	public ParseException(String documentName, int rowNumber, int columnNumber, String message, Throwable cause) {
		super(formatMessage(documentName, rowNumber, columnNumber, message), cause);
	}
	
	////////////////////
	// private methods
	
	/**
	 * Form a well-phrased error message reporting the given error in the given file!
	 * 
	 * @param fileName The name of the offending file.
	 * @param message A detailed description of the exception conditions.
	 * @return An informative error message.
	 */
	private static String formatMessage(String fileName, String message) {
		return String.format("Error reading '%s': %s!", fileName, message);
	}
	
	/**
	 * Form a well-phrased error message reporting the given error at the given position in the given file!
	 * 
	 * @param fileName The name of the offending file.
	 * @param lineNumber The line number in the file at which the exception occurred. Line numbers start at 1.
	 * @param message A detailed description of the exception conditions.
	 * @return An informative error message.
	 */
	private static String formatMessage(String fileName, int lineNumber, String message) {
		return String.format("Error in '%s', line %d: %s!", fileName, lineNumber, message);
	}
	
	/**
	 * Form a well-phrased error message reporting the given error at the given position in the given file!
	 * 
	 * @param documentName The name of the offending document.
	 * @param rowNumber The row number in the document at which the exception occurred. Row numbers start at 0, but are reported as starting from 1, i. e. the reported row number is the actual row number + 1.
	 * @param columnNumber The column number in the document at which the exception occurred. Column numbers start at 0, but are reported as starting from 1, i. e. the reported column number is the actual column number + 1.
	 * @param message A detailed description of the exception conditions.
	 * @return An informative error message.
	 */
	private static String formatMessage(String documentName, int rowNumber, int columnNumber, String message) {
		return String.format("Error in '%s', row %d, column %d: %s!", documentName, rowNumber+1, columnNumber+1, message);
	}
}
