/*
 Copyright (c) 2010--2011 Technische Universitt Darmstadt.
 
 Permission is hereby granted to use, copy, modify, and distribute this
 software, provided that this copyright notice remains intact.
 */

package gsi.rf.data.csv;

import java.util.Collections;
import java.util.Enumeration;
import java.util.LinkedList;
import java.util.ListIterator;

/**
 * This class represents a "character-separated values" (CSV) file.
 * 
 * A CSV file contains several records, separated by line-breaks.
 * Each record contains several fields, separated by a separator character
 * (usually one of comma (','), semicolon (';') or tabulator ('\t').
 * In general, each field is a character string.
 * However, some fields may be interpreted as numeric values, enumeration constants, or key-value pairs.
 * 
 * Internally, records are stored in a linked list.
 * Therefore, iterating over the records in order is faster than random access, especially for large documents.
 * 
 * @see CSVField
 * @see CSVReader
 * @see CSVRecord
 * 
 * @author Christopher Spies, TU Darmstadt
 * @since 2011-05-23
 * @version 1.5.0.0
 */
public class CSVDocument {

	////////////////////
	// private members
	
	/**
	 * A list of records within the file.
	 */
	private LinkedList<CSVRecord> records = new LinkedList<CSVRecord>();
	
	////////////////////
	// constructors
	
	/**
	 * Construct an empty document!
	 * 
	 * @since 2011-05-23
	 */
	public CSVDocument() {}
	
	////////////////////
	// public methods
	
	/**
	 * Add a new field to the last record of this document!
	 * 
	 * If the document is empty, a new record is created first.
	 * This function is offered for convenience and is a shortcut to {@link CSVRecord#addField(CSVField)}.
	 * 
	 * @param csvf The field to be added.
	 * @see CSVRecord#addField(CSVField)
	 * 
	 * @since 2011-05-25
	 */
	public void addField(CSVField csvf) {
		if (this.countRecords() <= 0) this.addRecord(new CSVRecord());
		this.records.getLast().addField(csvf);
	}
	
	/**
	 * Add a new record to the end of this document!
	 * 
	 * @param csvr The record to be added.
	 * 
	 * @since 2011-05-23
	 */
	public void addRecord(CSVRecord csvr) { this.records.addLast(csvr); }
	
	/**
	 * Clear this document!
	 * 
	 * @since 2011-06-30
	 */
	public void clear() { this.records.clear(); }
	
	/**
	 * Get the number of records in this document!
	 * 
	 * @return The number of records in the document.
	 * 
	 * @since 2011-05-23
	 */
	public int countRecords() { return this.records.size(); }
	
	/**
	 * Determine whether this document is equal to another object!
	 * 
	 * Two documents are considered equal if they have the same number of records
	 * and all of their records are equal.
	 * 
	 * @param other The other documents to compare to.
	 * @return TRUE If the other object is also a CSVDocument and equal to this one,
	 *         FALSE otherwise.
	 * @see CSVRecord#equals(Object)
	 * 
	 * @since 2011-06-10
	 */
	public boolean equals(Object other) {
		if (other == null) return false;
		if (other instanceof CSVDocument) {
			CSVDocument otherDocument = (CSVDocument)other;
			if (otherDocument.countRecords() != this.countRecords()) return false;
			Enumeration<CSVRecord> myRecords    = this.records();
			Enumeration<CSVRecord> otherRecords = otherDocument.records();
			while (myRecords.hasMoreElements() && otherRecords.hasMoreElements()) {
				if (otherRecords.nextElement().equals(myRecords.nextElement())) continue;
				return false;
			}
		}
		return true;
	}
	
	/**
	 * Get a specific field from this document!
	 * 
	 * This function is offered for convenience and is a shortcut to {@link CSVRecord#getField(int)}.
	 * 
	 * @param recordIndex The index of the record (index 0 means the first record in the document). Indices must be nonnegative.
	 * @param fieldIndex The index of the field (index 0 means the first field in the record). Indices must be nonnegative.
	 * @return The field with the given indices, or NULL if no such field exists.
	 * @throws IllegalArgumentException If either index is negative.
	 * @see CSVRecord#getField(int)
	 * 
	 * @since 2011-05-24
	 * @deprecated Iterate over the records in sequence!
	 */
	public CSVField getField(int recordIndex, int fieldIndex) throws IllegalArgumentException {
		CSVRecord csvr = this.getRecord(recordIndex);
		if (csvr == null) return null;
		return csvr.getField(fieldIndex);
	}
	
	/**
	 * Get the content of a specific field in this document!
	 * 
	 * This function is offered for convenience and is a shortcut to {@link CSVField#getContent()}.
	 * 
	 * @param recordIndex The index of the record (index 0 means the first record in the document). Indices must be nonnegative.
	 * @param fieldIndex The index of the field (index 0 means the first field in the record). Indices must be nonnegative.
	 * @return The content of the field with the given indices, or NULL if no such field exists.
	 * @throws IllegalArgumentException If either index is negative.
	 * @see CSVRecord#getFieldContent(int)
	 * @see CSVField#getContent()
	 * 
	 * @since 2011-05-24
	 * @deprecated Iterate over the records in sequence!
	 */
	public String getFieldContent(int recordIndex, int fieldIndex) throws IllegalArgumentException {
		CSVRecord csvr = this.getRecord(recordIndex);
		if (csvr == null) return null;
		return csvr.getFieldContent(fieldIndex);
	}
	
	/**
	 * Get a specific record from this document!
	 * 
	 * @param index The index of the record (index 0 means the first record in the document). Indices must be nonnegative.
	 * @return The record with the given index, or NULL if no such record exists.
	 * @throws IllegalArgumentException If the index is negative.
	 * 
	 * @since 2011-05-24
	 */
	public CSVRecord getRecord(int index) throws IllegalArgumentException {
		if (index < 0) throw new IllegalArgumentException("Index must be nonnegative!");
		try {
			return this.records.get(index);
		} catch (IndexOutOfBoundsException ioobx) {
			// an IndexOutOfBoundsException might be thrown if index is too large
			return null;
		}
	}
	
	/**
	 * Determine whether this record is empty!
	 * 
	 * A record is considered empty if it has zero records
	 * OR if all of its records are empty.
	 * 
	 * @return TRUE if the document is empty,
	 *         FALSE otherwise.
	 * @see CSVField
	 *
	 * @since 2011-05-25
	 */
	public boolean isEmpty() {
		if (this.records.isEmpty()) return true;
		for (CSVRecord aRecord : this.records) {
			if (aRecord == null) continue;
			if (!aRecord.isEmpty()) return false;
		}
		return true;
	}
	
	/**
	 * Get all records in this document!
	 * 
	 * @return An enumeration of all records in this document.
	 * 
	 * @since 2011-07-06
	 */
	public Enumeration<CSVRecord> records() { return Collections.enumeration(this.records); }
	
	/**
	 * Remove duplicate records from this document!
	 * 
	 * A record is considered to be a duplicate if it is equal to the record preceding it.
	 * 
	 * @see CSVRecord#equals(Object)
	 * 
	 * @since 2011-06-10
	 */
	public void removeDuplicateRecords() {
		CSVRecord previousRecord = null;
		CSVRecord nextRecord;
		
		ListIterator<CSVRecord> iterator = this.records.listIterator();
		while (iterator.hasNext()) {
			nextRecord = iterator.next();
			if (nextRecord.equals(previousRecord)) {
				iterator.remove();
				// PREVIOUSRECORD remains unchanged
			} else {
				previousRecord = nextRecord;
			}
		}
	}
	
	/**
	 * Set a specific record in this document!
	 * 
	 * @param index The index of the record (index 0 means the first record in the document). Indices must be nonnegative.
	 * @param csvr The new record.
	 * @throws IllegalArgumentException If the index is negative.
	 * 
	 * @since 2011-06-01
	 */
	public void setRecord(int index, CSVRecord csvr) throws IllegalArgumentException {
		if (index < 0) throw new IllegalArgumentException("Index must be nonnegative!");
		while (index >= this.countRecords()) this.records.add(null);
		this.records.set(index, csvr);
	}
	
	/**
	 * Set a specific field in this document!
	 * 
	 * This function is offered for convenience and is a shortcut to {@link CSVRecord#setField(int, CSVField)}.
	 * 
	 * @param recordIndex The index of the record (index 0 means the first record in the document). Indices must be nonnegative.
	 * @param fieldIndex The index of the field (index 0 means the first field in the record). Indices must be nonnegative.
	 * @param csvf The new field.
	 * @throws IllegalArgumentException If either index is negative.
	 * @see CSVRecord#setField(int)
	 * 
	 * @since 2011-06-01
	 * @deprecated Iterate over the records in sequence!
	 */
	public void setField(int recordIndex, int fieldIndex, CSVField csvf) throws IllegalArgumentException {
		CSVRecord csvr = this.getRecord(recordIndex);
		if (csvr == null) {
			csvr = new CSVRecord();
			this.setRecord(recordIndex, csvr);
		}
		csvr.setField(fieldIndex, csvf);
	}
	
	/**
	 * Set the content of a specific field in this document!
	 * 
	 * This function is offered for convenience and is a shortcut to {@link CSVField#setContent(String)}.
	 * 
	 * @param recordIndex The index of the record (index 0 means the first record in the document). Indices must be nonnegative.
	 * @param fieldIndex The index of the field (index 0 means the first field in the record). Indices must be nonnegative.
	 * @param content The new content.
	 * @throws IllegalArgumentException If either index is negative.
	 * @see CSVRecord#setFieldContent(int)
	 * @see CSVField#setContent()
	 * 
	 * @since 2011-05-25
	 * @deprecated Iterate over the records in sequence!
	 */
	public void setFieldContent(int recordIndex, int fieldIndex, String content) throws IllegalArgumentException {
		CSVRecord csvr = this.getRecord(recordIndex);
		if (csvr == null) {
			csvr = new CSVRecord();
			this.setRecord(recordIndex, csvr);
		}
		csvr.setFieldContent(fieldIndex, content);
	}
	
	/**
	 * Trim the records in this document!
	 * 
	 * Trimming means removing leading and trailing whitespace from all fields
	 * as well as removing empty fields from the end of each record
	 * and removing empty records from the end of the document.
	 * 
	 * @since 2011-06-09
	 */
	public void trim() {
		for (CSVRecord aRecord : this.records) aRecord.trim();
		
		ListIterator<CSVRecord> iterator = this.records.listIterator(this.records.size()); // obtain an iterator initially pointing to the last element in the list
		while (iterator.hasPrevious()) {
			CSVRecord lastRecord = iterator.previous();
			if (lastRecord.isEmpty()) iterator.remove();
			else break;
		}
	}
	
}
