/*
 Copyright (c) 2010--2011 Technische Universitt Darmstadt.
 
 Permission is hereby granted to use, copy, modify, and distribute this
 software, provided that this copyright notice remains intact.
 */

package gsi.rf.data.csv;

/**
 * This class represents a field within a "character-separated values" (CSV) file.
 * 
 * A CSV file contains several records (see {@link CSVRecord}).
 * Each record contains several fields, separated by a separator character
 * (usually one of comma (','), semicolon (';') or tabulator ('\t').
 * In general, each field is a character string.
 * However, some fields may be interpreted as numeric values, enumeration constants, or key-value pairs.
 * 
 * @see CSVDocument
 * @see CSVRecord
 * 
 * @author Christopher Spies, TU Darmstadt
 * @since 2011-05-23
 * @version 1.2.0.0
 */
public class CSVField {
	
	////////////////////
	// private members
	
	/**
	 * The content of this field.
	 */
	private String content = null;
	
	////////////////////
	// constructors
	
	/**
	 * Construct an empty field!
	 * 
	 * @since 2011-05-23
	 */
	public CSVField() {}
	
	/**
	 * Construct a field with the given content!
	 * 
	 * @param content The content.
	 * 
	 * @since 2011-05-23
	 */
	public CSVField(String content) {
		this();
		this.setContent(content);
	}

	////////////////////
	// public methods
	
	/**
	 * Determine whether this field is equal to another object!
	 * 
	 * Two fields are considered equal if their content is equal.
	 * 
	 * @param other The other field to compare to.
	 * @return TRUE If the other object is also a CSVField and equal to this one,
	 *         FALSE otherwise.
	 * @see String#equals(Object)
	 * 
	 * @since 2011-06-10
	 */
	public boolean equals(Object other) {
		if (other == null) return false;
		if (other instanceof CSVField) {
			CSVField otherField = (CSVField)other;
			return this.getContent().equals(otherField.getContent());
		} else return false;
	}
	
	/**
	 * Get the content of this field!
	 * 
	 * @return The content of this field.
	 * 
	 * @since 2011-05-23
	 */
	public String getContent() {
		return this.content;
	}
	
	/**
	 * Determine whether this field is empty!
	 * 
	 * A field is considered empty if the content is NULL or the length of its content is zero.
	 * A field consisting entirely of whitespace is NOT considered to be empty.
	 * Note that leading and trailing whitespace might be removed by the reader or writer.
	 * 
	 * @return TRUE if the field is empty,
	 *         FALSE otherwise.
	 * @see CSVReader
	 * @see CSVWriter
	 *
	 * @since 2011-05-25
	 */
	public boolean isEmpty() {
		if (this.content == null) return true;
		if (this.content.isEmpty()) return true;
		return false;
	}
	
	/**
	 * Set the content of this field!
	 * 
	 * @param content The new content.
	 * 
	 * @since 2011-05-23
	 */
	public void setContent(String content) {
		this.content = content;
	}
	
	/**
	 * Trim the content of this field!
	 * 
	 * Trimming means removing leading and trailing whitespace.
	 * 
	 * @since 2011-06-09
	 */
	public void trim() {
		this.content = this.content.trim();
	}

}
