/*
 Copyright (c) 2010--2012 Technische Universitt Darmstadt.
 
 Permission is hereby granted to use, copy, modify, and distribute this
 software, provided that this copyright notice remains intact.
 */

package gsi.rf.data.csv;

import gsi.rf.data.ParseException;
import gsi.rf.data.csv.parser.CSVParser;
import gsi.rf.data.csv.parser.CSVScanner;
import gsi.rf.data.csv.parser.IllegalCharacterException;
import gsi.rf.data.csv.parser.Token;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.Set;

/**
 * This class represents a reader for "character-separated values" (CSV) files.
 * 
 * It might appear unnecessary to have a dedicated reader class for CSV files,
 * since the {@link CSVDocument} class could just as well implement its own read method.
 * However, a separate reader class leaves room for future extensions
 * and relieves the {@link CSVDocument} class of the overhead of the additional methods
 * as well as the complex internal state of the CSV parser.
 * 
 * A CSV file contains several records, separated by line-breaks.
 * Each record contains several fields, separated by a separator character
 * (usually one of comma (','), semicolon (';') or tabulator ('\t').
 * In general, each field is a character string.
 * However, some fields may be interpreted as numeric values, enumeration constants, or key-value pairs.
 * 
 * The grammar of a CSV file is officially defined by RFC 4180 as follows:
 * file        ::= record { linebreak record } .
 * record      ::= field { separator field } .
 * field       ::= escaped | non-escaped .
 * escaped     ::= quote { textdata | separator | linebreak | 2quotes } quote .
 * non-escaped ::= { textdata }.
 * linebreak   ::= '\r' \'n' .
 * separator   ::= ',' .
 * quote       ::= '"' .
 * 2quotes     ::= '"' '"' .
 * textdata    ::= < any character except '\n', '\r', ',' and '"' >
 *
 * In order to conform to GSI's standard for data analysis file formats for RF applications
 * (at present, draft revision 0.60 from May 24th, 2011),
 * this parser
 * - accepts any combination of carriage return ('\r', 0x0D) and line feed ('\n', 0x0A) as a line break,
 * - can be configured to accept not only commas (','), but also semicolons (';') and tabulators ('\t') as field separators,
 * - accepts leading and trailing whitespace in fields and
 * - can handle quotation marks as parts of non-escaped fields.
 * 
 * @see CSVDocument
 * @see CSVField
 * @see CSVRecord
 * 
 * @author Christopher Spies, TU Darmstadt
 * @since 2011-05-23
 * @version 1.0.1.1
 */
public class CSVReader {

	////////////////////
	// private constants
	
	////////////////////
	// public constants
	
   /**
	 * The set of default field separators.
	 * 
	 * This field contains only the comma character (',').
	 * It is provided as a public constant for convenience.
	 * 
	 * @see CSVScanner#DEFAULT_FIELD_SEPARATOR_SET
	 * 
	 * @since 2011-05-23
	 */
	public final static Set<Character> DEFAULT_FIELD_SEPARATOR_SET = CSVScanner.DEFAULT_FIELD_SEPARATOR_SET;
	
	/**
	 * The extended set of field separators.
	 * 
	 * This field contains the comma (','), the semicolon (';') and the tabulator ('\t') character.
	 * It is provided as a public constant for convenience.
	 * 
	 * @see CSVScanner#EXTENDED_FIELD_SEPARATOR_SET
	 * 
	 * @since 2011-05-23
	 */
	public final static Set<Character> EXTENDED_FIELD_SEPARATOR_SET = CSVScanner.EXTENDED_FIELD_SEPARATOR_SET;
	
	/**
	 * The set of characters that must never be field separators.
	 * 
	 * This field contains the carriage return ('\r'), line feed ('\n'), blank space (' ') and quotation mark ('"') characters.
	 * It is provided as a public constant for convenience.
	 * 
	 * @see CSVScanner#UNSUITABLE_FIELD_SEPARATOR_SET
	 * 
	 * @since 2011-05-24
	 */
	public final static Set<Character> UNSUITABLE_FIELD_SEPARATOR_SET = CSVScanner.UNSUITABLE_FIELD_SEPARATOR_SET;
	
	////////////////////
	// private members

	/**
	 * The parser used by this reader.
	 */
	private CSVParser parser;  
	
	/**
	 * The scanner used by this reader.
	 */
	private CSVScanner scanner;
	
	////////////////////
	// constructors
	
	/**
	 * Construct a new reader!
	 * 
	 * A parser object created using the default constructor strictly adheres to RFC 4180,
	 * and allows only commas as field separators.
	 * 
	 * @since 2011-05-23
	 */
	public CSVReader() {
		this.scanner = new CSVScanner();
		this.parser  = new CSVParser();
	}
	
	////////////////////
	// public methods
	
	/**
	 * Read and parse a CSV file!
	 * 
	 * @param f The file to be read.
	 * @return A CSVDocument containing the data parsed from the file.
	 * @throws IOException If any error occurs while trying to locate, open and read the input file.
	 * @throws ParseException If any error occurs while trying to parse the input file.
	 * 
	 * @since 2011-05-23
	 */
	public CSVDocument read(File f) throws IOException, ParseException {
		if (f == null) throw new FileNotFoundException("File is NULL");
		FileReader reader;
		int aCharacter;
		Token aToken;
		
		this.scanner.begin(f.getName());
		this.parser.begin(f.getName());
		
		// iterate over all characters in the file:
		try {
			reader = new FileReader(f);
			
			while (true) {
				aCharacter = reader.read();
				if (aCharacter == -1) break; // end of file
				this.scanner.advanceState((char)aCharacter);
			}
			this.scanner.terminate();
		} catch (IllegalCharacterException icx) {
			throw new ParseException(f.getName(), icx);
		}
		
		// iterate over all tokens created by the scanner:
		while (this.scanner.hasToken()) {
			aToken = this.scanner.consumeToken();
			this.parser.advanceState(aToken);
		}
		this.parser.terminate();

		return this.parser.getDocument();
	}
	
	/**
	 * Define the set of acceptable field separators!
	 * 
	 * @param separators A set of characters to be accepted as field separators.
	 * @throws IllegalArgumentException If the desired set contains characters that are already used otherwise (e. g. as line breaks).
	 * 
	 * @since 2011-05-23
	 */
	public void setAcceptableFieldSeparators(Set<Character> separators) throws IllegalArgumentException { this.scanner.setAcceptableFieldSeparators(separators); }
	
	/**
	 * Turn strict parsing off, i. e. enable lenient parsing!
	 * 
	 * Lenient parsing means trying to continue parsing even if a violation of RFC 4180 is being detected, such as
	 * - line breaks marked by CR only, LF only or LF+CR instead of CR+LF
	 * - un-escaped quotation marks inside escaped fields.
	 * 
	 * @since 2011-05-23
	 */
	public void turnStrictParsingOff() {
		this.scanner.turnStrictScanningOff();
		this.parser.turnStrictParsingOff();
	}
	
	/**
	 * Turn strict parsing on, i. e. strictly adhere to RFC 4180!
	 * 
	 * Strict parsing means throwing an exception every time a violation of RFC 4180 is being detected, such as
	 * - line breaks marked by CR only, LF only or LF+CR instead of CR+LF
	 * - un-escaped quotation marks inside escaped fields.
	 * 
	 * @since 2011-05-23 
	 */
	public void turnStrictParsingOn() {
		this.scanner.turnStrictScanningOn();
		this.parser.turnStrictParsingOn();
	}
	
	////////////////////
	// private methods
	
	////////////////////
	// local sub-classes
	
}
