/*
 Copyright (c) 2010--2011 Technische Universitt Darmstadt.
 
 Permission is hereby granted to use, copy, modify, and distribute this
 software, provided that this copyright notice remains intact.
 */

package gsi.rf.data.csv;

import java.util.ArrayList;
import java.util.List;
import java.util.ListIterator;

/**
 * This class represents a record within a "character-separated values" (CSV) file.
 * 
 * A CSV file contains several records, separated by line-breaks.
 * Each record contains several fields, separated by a separator character
 * (usually one of comma (','), semicolon (';') or tabulator ('\t').
 * In general, each field is a character string.
 * However, some fields may be interpreted as numeric values, enumeration constants, or key-value pairs.
 * 
 * @see CSVDocument
 * @see CSVField
 * 
 * @author Christopher Spies, TU Darmstadt
 * @since 2011-05-23
 * @version 1.2.0.0
 */
public class CSVRecord {

	////////////////////
	// private members
	
	/**
	 * A list of fields within the record.
	 */
	private List<CSVField> fields = new ArrayList<CSVField>();
	
	////////////////////
	// constructors
	
	/**
	 * Construct an empty record!
	 * 
	 * @since 2011-05-23
	 */
	public CSVRecord() {}
	
	////////////////////
	// public methods
	
	/**
	 * Add a new field to the end of this record!
	 * 
	 * @param aField The field to be added.
	 * 
	 * @since 2011-05-23
	 */
	public void addField(CSVField csvf) { this.fields.add(csvf); }
	
	/**
	 * Get the number of fields in this record!
	 * 
	 * @return The number of fields in the CSVRecord.
	 * 
	 * @since 2011-05-23
	 */
	public int countFields() { return this.fields.size(); }
	
	/**
	 * Determine whether this record is equal to another object!
	 * 
	 * Two records are considered equal if they have the same number of fields
	 * and all of their fields are equal.
	 * 
	 * @param other The other record to compare to.
	 * @return TRUE If the other object is also a CSVRecord and equal to this one,
	 *         FALSE otherwise.
	 * @see CSVField#equals(Object)
	 * 
	 * @since 2011-06-10
	 */
	public boolean equals(Object other) {
		if (other == null) return false;
		if (other instanceof CSVRecord) {
			CSVRecord otherRecord = (CSVRecord)other;
			int numberOfFields = this.countFields();
			if (otherRecord.countFields() != numberOfFields) return false;
			for (int fieldIndex=0; fieldIndex<numberOfFields; ++fieldIndex) {
				CSVField myField    = this.getField(fieldIndex);
				CSVField otherField = otherRecord.getField(fieldIndex);
				if ((myField == null) && (otherField != null)) return false;
				if (!myField.equals(otherField)) return false;
			}
		}
		return true;
	}
	
	/**
	 * Get a specific field from this record!
	 * 
	 * @param index The index of the field (index 0 means the first field in the record). Indices must be nonnegative.
	 * @return The field with the given index, or NULL if no such field exists.
	 * @throws IllegalArgumentException If the index is negative.
	 * 
	 * @since 2011-05-24
	 */
	public CSVField getField(int index) throws IllegalArgumentException {
		if (index < 0) throw new IllegalArgumentException("Index must be nonnegative!");
		try {
			return this.fields.get(index);
		} catch (IndexOutOfBoundsException ioobx) {
			// an IndexOutOfBoundsException might be thrown if index is too large
			return null;
		}
	}
	
	/**
	 * Get the content of a specific field from this record!
	 * 
	 * This function is offered for convenience and is a shortcut for {@link CSVField#getContent()}.
	 * 
	 * @param index The index of the field (index 0 means the first field in the record). Indices must be nonnegative.
	 * @return The content of the field with the given index, or NULL if no such field exists.
	 * @see CSVField#getContent()
	 * @throws IllegalArgumentException If the index is negative.
	 * 
	 * @since 2011-05-24
	 */
	public String getFieldContent(int index) throws IllegalArgumentException {
		CSVField csvf = this.getField(index);
		if (csvf == null) return null;
		return csvf.getContent();
	}
	
	/**
	 * Determine whether this record is empty!
	 * 
	 * A record is considered empty if it has zero fields
	 * OR if all of its fields are empty.
	 * 
	 * @return TRUE if the record is empty,
	 *         FALSE otherwise.
	 * @see CSVField
	 *
	 * @since 2011-05-25
	 */
	public boolean isEmpty() {
		if (this.fields.isEmpty()) return true;
		for (CSVField aField : this.fields) {
			if (aField == null) continue;
			if (!aField.isEmpty()) return false;
		}
		return true;
	}
	
	/**
	 * Set a specific field in this record!
	 * 
	 * @param index The index of the field (index 0 means the first field in the record). Indices must be nonnegative.
	 * @param csvf The new field.
	 * @throws IllegalArgumentException If the index is negative.
	 * @see #setFieldContent(int, String)
	 * 
	 * @since 2011-06-01
	 */
	public void setField(int index, CSVField csvf) throws IllegalArgumentException {
		if (index < 0) throw new IllegalArgumentException("Index must be nonnegative!");
		while (index >= this.countFields()) this.fields.add(null);
		this.fields.set(index, csvf);
	}
	
	/**
	 * Set the content of a specific field from this record!
	 * 
	 * This function is offered for convenience and is a shortcut for {@link CSVField#setContent(String)}.
	 * 
	 * @param index The index of the field (index 0 means the first field in the record). Indices must be nonnegative.
	 * @param content The new content.
	 * @throws IllegalArgumentException If the index is negative.
	 * @see CSVField#getContent()
	 * 
	 * @since 2011-05-25
	 */
	public void setFieldContent(int index, String content) throws IllegalArgumentException {
		CSVField csvf = this.getField(index);
		if (csvf == null) {
			csvf = new CSVField();
			this.setField(index, csvf);
		}
		csvf.setContent(content);
	}
	
	/**
	 * Trim the fields in this record!
	 * 
	 * Trimming means removing leading and trailing whitespace from all fields
	 * as well as removing empty fields from the end of the record.
	 * 
	 * @since 2011-06-09
	 */
	public void trim() {
		for (CSVField aField : this.fields) aField.trim();
		
		ListIterator<CSVField> iterator = this.fields.listIterator(this.fields.size()); // obtain an iterator initially pointing to the last element in the list
		while (iterator.hasPrevious()) {
			CSVField lastField = iterator.previous();
			if (lastField.isEmpty()) iterator.remove();
			else break;
		}
	}
	
}
