/*
 Copyright (c) 2010--2011 Technische Universitt Darmstadt.
 
 Permission is hereby granted to use, copy, modify, and distribute this
 software, provided that this copyright notice remains intact.
 */

package gsi.rf.data.csv;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Enumeration;

/**
 * This class represents a writer for "character-separated values" (CSV) files.
 * 
 * It might appear unnecessary to have a dedicated writer class for CSV files,
 * since the {@link CSVDocument} class could just as well implement its own write method.
 * However, a separate writer class leaves room for future extensions
 * and relieves the {@link CSVDocument} class of the overhead of the additional methods.
 * 
 * A CSV file contains several records, separated by line-breaks ('\r' '\n').
 * Each record contains several fields, separated by commas (',').
 * In general, each field is a character string.
 * However, some fields may be interpreted as numeric values, enumeration constants, or key-value pairs.
 * 
 * @see CSVDocument
 * @see CSVField
 * @see CSVRecord
 * 
 * @author Christopher Spies, TU Darmstadt
 * @since 2011-05-25
 * @version 1.1.0.0
 */
public class CSVWriter {

	////////////////////
	// private constants
	
	/**
	 * The carriage-return character.
	 * 
	 * This character is accepted as a line break, either by itself or in any combination with the line-feed character.
	 * 
	 * @see #LINE_FEED
	 */
	private final static Character CARRIAGE_RETURN = (char)0x0D;
	
	/**
	 * The comma character.
	 * 
	 * This character is the default field separator.
	 */
	private final static Character COMMA = ',';
	
	/**
	 * The field separator.
	 */
	private final static String FIELD_SEPARATOR = COMMA.toString();
	
	/**
	 * The line-feed character.
	 * 
	 * This character is accepted as a line break, either by itself or in any combination with the carriage-return character.
	 * 
	 * @see #CARRIAGE_RETURN
	 */
	private final static Character LINE_FEED = (char)0x0A;

	/**
	 * Double quotation marks.
	 * 
	 * This character is the default escape character.
	 */
	private final static Character QUOTATION_MARK = '"';
	
	/**
	 * The record separator.
	 */
	private final static String RECORD_SEPARATOR = CARRIAGE_RETURN.toString().concat(LINE_FEED.toString());
	
	/**
	 * The delimiter used to mark the beginning and end of escaped fields: The double quotation mark character ("\"").
	 * 
	 * @see #QUOTATION_MARK
	 */
	private final static String STRING_DELIMITER = QUOTATION_MARK.toString();
	
	/**
	 * The escape sequence used when the delimiter appears inside an escaped field: Dual double quotation marks ("\"\"").
	 * 
	 * @see #QUOTATION_MARK
	 * @see #STRING_DELIMITER
	 */
	private final static String STRING_DELIMITER_ESCAPE_SEQUENCE = STRING_DELIMITER.concat(STRING_DELIMITER);
	
	////////////////////
	// private members
	
	/**
	 * Indicates whether leading and trailing whitespace should be removed before writing.
	 * TRUE  = Trim whitespace,
	 * FALSE = Retain original character string.
	 */
	private boolean doTrim = false;

	////////////////////
	// constructors
	
	/**
	 * Construct a new writer!
	 * 
	 * @since 2011-05-25
	 */
	public CSVWriter() {
	}
	
	////////////////////
	// public methods
	
	/**
	 * Turn output trimming off, i. e. don't remove leading or trailing whitespace from fields!
	 * 
	 * @since 2011-05-26
	 */
	public void turnTrimmingOff() { this.doTrim = false; }
	
	/**
	 * Turn output trimming on, i. e. remove leading or trailing whitespace from fields!
	 * 
	 * @since 2011-05-26
	 */
	public void turnTrimmingOn() { this.doTrim = true; }
	
	/**
	 * Write a CSV file!
	 * 
	 * @param csvd The document containing the data to be written.
	 * @param f The file to be written to.
	 * @throws IOException If any error occurs while trying to locate, create, open or write to the output file.
	 * 
	 * @since 2011-05-25
	 */
	public void write(CSVDocument csvd, File f) throws IOException { this.write(csvd, f, false); }
	
	/**
	 * Write or append to a CSV file!
	 * 
	 * @param csvd The document containing the data to be written.
	 * @param f The file to be written to.
	 * @param append TRUE if the document is to be appended to the file,
	 *               FALSE if the file is to be overwritten if it exists already.
	 * @throws IOException If any error occurs while trying to locate, create, open or write to the output file.
	 * 
	 * @since 2011-07-07
	 */
	public void write(CSVDocument csvd, File f, boolean append) throws IOException {
		FileWriter writer;
		
		writer = new FileWriter(f, append);
		if (append) writer.write(RECORD_SEPARATOR);
		this.writeDocument(csvd, writer);
		writer.close();
	}
	
	////////////////////
	// private methods
	
	/**
	 * Escape the given content!
	 * 
	 * Here, "escaping" content means to put a string delimiter in front of and behind the original content
	 * and replacing any occurrence of the string delimiter in the original content by an escape sequence.
	 * 
	 * @param content The content to be escaped.
	 * @return The escaped content.
	 * 
	 * @see #STRING_DELIMITER
	 * @see #STRING_DELIMITER_ESCAPE_SEQUENCE
	 */
	private static String escape(String content) {
		StringBuilder escapedContent = new StringBuilder();
		escapedContent.append(STRING_DELIMITER);
		escapedContent.append(content.replace(STRING_DELIMITER, STRING_DELIMITER_ESCAPE_SEQUENCE));
		escapedContent.append(STRING_DELIMITER);
		return escapedContent.toString();
	}

	/**
	 * Write the given document to a file using the given FileWriter!
	 * 
	 * @param csvd The document to be written.
	 * @param writer The FileWriter to use.
	 * @throws IOException If any error occurs while trying to write to the output file.
	 */
	private void writeDocument(CSVDocument csvd, FileWriter writer) throws IOException {
		Enumeration<CSVRecord> records = csvd.records();
		for (int recordIndex=0; records.hasMoreElements(); ++recordIndex) {
			CSVRecord aRecord = records.nextElement();
			if (recordIndex > 0) writer.write(RECORD_SEPARATOR);
			this.writeRecord(aRecord, writer);
		}
	}
	
	/**
	 * Write the given field to a file using the given FileWriter!
	 * 
	 * The field's content may have to be escaped first.
	 * 
	 * @param csvf The field to be written.
	 * @param writer The FileWriter to use.
	 * @throws IOException If any error occurs while trying to write to the output file.
	 * @see #escape(String)
	 */
	private void writeField(CSVField csvf, FileWriter writer) throws IOException {
		String content = csvf.getContent();
		if (this.doTrim)content = content.trim(); // remove leading and trailing whitespace
		if ((content.indexOf(CARRIAGE_RETURN) >= 0)
		 || (content.indexOf(COMMA) >= 0)
		 || (content.indexOf(LINE_FEED) >= 0)
		 || (content.indexOf(QUOTATION_MARK) >= 0))
		{
			// any field containing line breaks, field separators or escape characters has to be escaped
			// (enclosed in quotation marks)
			content = escape(content);
		}
		writer.write(content);
	}
	
	/**
	 * Write the given record to a file using the given FileWriter!
	 * 
	 * @param csvr The record to be written.
	 * @param writer The FileWriter to use.
	 * @throws IOException If any error occurs while trying to write to the output file.
	 */
	private void writeRecord(CSVRecord csvr, FileWriter writer) throws IOException {
		int fieldCount = csvr.countFields();
		for (int i=0; i<fieldCount; ++i) {
			CSVField csvf = csvr.getField(i);
			if (i > 0) writer.write(FIELD_SEPARATOR);
			this.writeField(csvf, writer);
		}
	}
}
