/*
 Copyright (c) 2010--2012 Technische Universitt Darmstadt.
 
 Permission is hereby granted to use, copy, modify, and distribute this
 software, provided that this copyright notice remains intact.
 */

package gsi.rf.data.csv.parser;

import gsi.rf.data.ParseException;
import gsi.rf.data.csv.CSVDocument;
import gsi.rf.data.csv.CSVField;
import gsi.rf.data.csv.CSVRecord;

/**
 * This class represents a parser for "character-separated values" (CSV) files.
 * 
 * The grammar of a CSV file is officially defined by RFC 4180 as follows:
 * file        ::= record { line-break record } .
 * record      ::= field { separator field } .
 * field       ::= escaped | non-escaped .
 * escaped     ::= quote { text-data | separator | cr | lf | line-break | 2quotes } quote .
 * non-escaped ::= { text-data }.
 *
 * In order to conform to GSI's standard for data analysis file formats for RF applications
 * (at present, draft revision 0.80 from May 31st, 2011),
 * this parser
 * - can handle quotation marks in non-escaped fields and
 * - can handle leading and trailing whitespace in fields.
 *  
 * @author Christopher Spies, TU Darmstadt
 * @since 2011-09-12
 * @version 0.1
 */
public class CSVParser {

	////////////////////
	// private constants
	
	/**
	 * Double quotation marks.
	 * 
	 * This character is the default escape character.
	 */
	private final static Character QUOTATION_MARK = '"';
	
	////////////////////
	// private members
	
	/**
	 * A buffer containing the text that makes up the current field.
	 */
	private StringBuilder buffer;
	
	/**
	 * The document resulting from the parsing process.
	 */
	private CSVDocument currentDocument;
	
	/**
	 * The record which is currently being parsed.
	 */
	private CSVRecord currentRecord;
	
	/**
	 * The name of the file being scanned.
	 * 
	 * This information is required in error messages only.
	 */
	private String fileName;
	
	/**
	 * The current line number.
	 */
	private int lineNumber;
	
	/**
	 * Indicates whether parsing should strictly adhere to RFC 4180 or not.
	 * TRUE  = Adhere to RFC 4180,
	 * FALSE = Be lenient.
	 */
	private boolean isStrict = true;
	
	/**
	 * The internal state of the parser.
	 */
	private ParserState state;
	
	////////////////////
	// constructors
	
	/**
	 * Create a new parser!
	 * 
	 * @since 2011-09-11
	 */
	public CSVParser() {
		this.buffer = new StringBuilder();
		this.isStrict = true;
		this.state = ParserState.END_OF_FILE;
	}
	
	////////////////////
	// public methods
	
	/**
	 * Advance the state of the parser based on the next token!
	 * 
	 * @param nextToken The next token.
	 * @throws ParseException If the token violates the grammar.
	 * 
	 * @see #reset()
	 * @see #terminate()
	 * 
	 * @since 2011-09-02
	 */
	public void advanceState(Token nextToken) throws ParseException {
		ParserState nextState = this.state;
		TokenKind kind = nextToken.getKind();
		
		this.fileName   = nextToken.getFileName();
		this.lineNumber = nextToken.getLineNumber();
		
		switch (this.state) {
		case INITIAL_STATE:
			switch (kind) {
			case WHITESPACE: // whitespace is treated like text-data in strict mode and ignored otherwise
				if (!this.isStrict) break;
				// else fall through
			case TEXT_DATA:       this.buffer.append(nextToken.getText()); nextState = ParserState.NONESCAPED_FIELD; break;
			case FIELD_SEPARATOR: this.leaveField();                                                                 break;
			case LINE_BREAK:      this.leaveRecord();                                                                break;
			case QUOTATION_MARK:                                           nextState = ParserState.ESCAPED_FIELD;    break;
			case CARRIAGE_RETURN: // fall through
			case LINE_FEED:       // fall through
			case ESCAPE_SEQUENCE: this.reportUnexpected(nextToken);
			}
			break;
		case ESCAPED_FIELD:
			switch (kind) {
			case WHITESPACE:      // fall through
			case TEXT_DATA:       // fall through
			case FIELD_SEPARATOR: // fall through
			case CARRIAGE_RETURN: // fall through
			case LINE_FEED:       // fall through
			case LINE_BREAK:      this.buffer.append(nextToken.getText());                                              break;
			case QUOTATION_MARK:                                           nextState = ParserState.TRAILING_WHITESPACE; break;
			case ESCAPE_SEQUENCE: this.buffer.append(QUOTATION_MARK);                                                   break;
			}
			break;
		case NONESCAPED_FIELD:
			switch (kind) {
			case WHITESPACE:      // fall through
			case TEXT_DATA:       this.buffer.append(nextToken.getText());                                           break;
			case FIELD_SEPARATOR: this.leaveField();                          nextState = ParserState.INITIAL_STATE; break;
			case CARRIAGE_RETURN: // fall through
			case LINE_FEED:       this.reportUnexpected(nextToken);                                                  break;
			case LINE_BREAK:      this.leaveRecord();                         nextState = ParserState.INITIAL_STATE; break;
			case QUOTATION_MARK: // fall through
			case ESCAPE_SEQUENCE: // in strict mode, quotation marks inside a non-escaped field cannot be accepted
				if (this.isStrict) { this.reportUnexpected(nextToken);                                               break; }
				else               { this.buffer.append(nextToken.getText());                                        break; }
			}
			break;
		case TRAILING_WHITESPACE:
			switch (kind) {
			case WHITESPACE:                                                                               break;
			case FIELD_SEPARATOR: this.leaveField();                nextState = ParserState.INITIAL_STATE; break;
			case LINE_BREAK:      this.leaveRecord();               nextState = ParserState.INITIAL_STATE; break;
			case CARRIAGE_RETURN: // fall through
			case LINE_FEED:       // fall through
			case QUOTATION_MARK:  // fall through
			case ESCAPE_SEQUENCE: this.reportUnexpected(nextToken);                                        break;
			}
			break;
		case END_OF_FILE:
			throw new IllegalStateException("Input had been terminated!");
		}
		
		this.state = nextState;
	}
	
	/**
	 * Reset this scanner to its initial state and begin scanning!
	 * 
	 * @param fileName The name of the file to be scanned. This name will only be used in error messages.
	 * 
	 * @see #advance(char)
	 * @see #terminate()
	 * 
	 * @since 2011-09-02
	 */
	public void begin(String fileName) {
		this.buffer.setLength(0);
		this.fileName = fileName;
		this.lineNumber = 1;
		this.state = ParserState.INITIAL_STATE;
		this.currentDocument = new CSVDocument();
		this.currentRecord = new CSVRecord();
	}
	
	/**
	 * Get the document created by the parser!
	 * 
	 * @return The parsed document.
	 * 
	 * @since 2011-09-12
	 */
	public CSVDocument getDocument() { return this.currentDocument; }
	
	/**
	 * Terminate the input!
	 * 
	 * @return TRUE if at least one token was produced, FALSE otherwise.
	 * @throws ParseException If more tokens are expected.
	 * 
	 * @see #reset()
	 * @see #advance(Token)
	 * 
	 * @since 2011-09-12
	 */
	public boolean terminate() throws ParseException {
		boolean gotNewToken = false;
		
		switch (this.state) {
		case ESCAPED_FIELD:       throw new ParseException(this.fileName, this.lineNumber, "Unexpected end-of-file");
		case NONESCAPED_FIELD:    this.leaveField(); this.leaveRecord(); break;
		case TRAILING_WHITESPACE: this.leaveField(); this.leaveRecord(); break;
		default: // do nothing
		}
		
		this.state = ParserState.END_OF_FILE;
		return gotNewToken;
	}
	
	/**
	 * Turn strict parsing off, i. e. enable lenient parsing!
	 * 
	 * Lenient parsing means trying to continue parsing even if a violation of RFC 4180 is being detected, such as
	 * - line breaks marked by CR only, LF only or LF+CR instead of CR+LF
	 * - un-escaped quotation marks inside escaped fields.
	 * 
	 * @since 2011-05-23
	 */
	public void turnStrictParsingOff() { this.isStrict = false; }
	
	/**
	 * Turn strict parsing on, i. e. strictly adhere to RFC 4180!
	 * 
	 * Strict parsing means throwing an exception every time a violation of RFC 4180 is being detected, such as
	 * - line breaks marked by CR only, LF only or LF+CR instead of CR+LF
	 * - un-escaped quotation marks inside escaped fields.
	 * 
	 * @since 2011-05-23 
	 */
	public void turnStrictParsingOn() { this.isStrict = true; }
	
	////////////////////
	// private methods
	
	/**
	 * Leave the current field!
	 * 
	 * This method creates a new field from the content of the buffer,
	 * adds that field to the current record
	 * and empties the buffer afterwards.
	 */
	private void leaveField() {
		String fieldContent = this.buffer.toString();
		if (!this.isStrict) fieldContent = fieldContent.trim(); // in strict mode, all whitespace must be retained; otherwise, leading and trailing whitespace should be removed
		this.currentRecord.addField(new CSVField(fieldContent));
		this.buffer.setLength(0); // truncate the buffer
	}
	
	/**
	 * Leave the current record!
	 * 
	 * This method creates a new field from the content of the buffer,
	 * adds that field to the current record,
	 * adds the current record to the current document
	 * and empties the buffer afterwards.
	 */
	private void leaveRecord() {
		this.leaveField();
		this.currentDocument.addRecord(this.currentRecord);
		this.currentRecord = new CSVRecord();
	}
	
	/**
	 * Report an error due to an unexpected token!
	 * 
	 * @param offender The offending token.
	 * @throws ParseException
	 */
	private void reportUnexpected(Token offender) throws ParseException {
		throw new ParseException(offender.getFileName(), offender.getLineNumber(), String.format("Unexpected %s (\"%s\")", offender.getKind().getDescription(), offender.getText()));
	}
	
	////////////////////
	// local sub-classes
	
	/**
     * An enumeration of the parser states.
     * 
     * @since 2011-09-12
     */
    private enum ParserState {
        INITIAL_STATE,       // initial state before any data has been encountered
        ESCAPED_FIELD,       // inside an escaped field
        NONESCAPED_FIELD,    // inside a non-escaped field
        TRAILING_WHITESPACE, // whitespace following an escaped field
        END_OF_FILE          // end of file
    }
}
