/*
 Copyright (c) 2010--2012 Technische Universitt Darmstadt.
 
 Permission is hereby granted to use, copy, modify, and distribute this
 software, provided that this copyright notice remains intact.
 */

package gsi.rf.data.csv.parser;

import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Set;

/**
 * This class represents a lexical scanner for "character-separated values" (CSV) files.
 * 
 * The lexemes of a CSV file are officially defined by RFC 4180 as follows:
 * line-break ::= '\r' '\n' .
 * cr         ::= '\r' .
 * lf         ::= '\n' .
 * separator  ::= ',' .
 * quote      ::= '"' .
 * 2quotes    ::= '"' '"' .
 *
 * In order to conform to GSI's standard for data analysis file formats for RF applications
 * (at present, draft revision 0.80 from May 31st, 2011),
 * this scanner
 * - can be configured to accept accept any combination of carriage return ('\r', 0x0D) and line feed ('\n', 0x0A) as a line break,
 * - can be configured to accept not only commas (','), but also semicolons (';') and tabulators ('\t') as field separators and
 * - can be configured to distinguish whitespace from text-data.
 *  
 * @author Christopher Spies, TU Darmstadt
 * @since 2011-09-09
 * @version 0.1
 */
public class CSVScanner {
	
	////////////////////
	// private constants
	
	/**
	 * The comma character.
	 * 
	 * This character is the default field separator.
	 */
	private final static Character COMMA = ',';
	
	/**
	 * The carriage-return character.
	 * 
	 * This character is accepted as a line break, either by itself or in any combination with the line-feed character.
	 * 
	 * @see #LINE_FEED
	 */
	private final static Character CARRIAGE_RETURN = (char)0x0D;
	
	/**
	 * The horizontal tabulator character.
	 * 
	 * This character may be used as an alternative field separator; it is recognized as white space otherwise.
	 */
	private final static Character HORIZONTAL_TABULATOR = (char)0x09;
	
	/**
	 * The line-feed character.
	 * 
	 * This character is accepted as a line break, either by itself or in any combination with the carriage-return character.
	 * 
	 * @see #CARRIAGE_RETURN
	 */
	private final static Character LINE_FEED = (char)0x0A;
	
	/**
	 * Double quotation marks.
	 * 
	 * This character is the default escape character.
	 */
	private final static Character QUOTATION_MARK = '"';
	
	/**
	 * The first character in the range of characters acceptable as text data according to RFC 4180 (' ' = 0x20).
	 * 
	 * @see #RFC_4180_TEXT_DATA_RANGE_TO
	 */
	private final static Character RFC_4180_TEXT_DATA_RANGE_FROM = ' ';
	
	/**
	 * The last character in the range of characters acceptable as text data according to RFC 4180 ('~' = 0x7E).
	 * 
	 * @see #RFC_4180_TEXT_DATA_RANGE_FROM
	 */
	private final static Character RFC_4180_TEXT_DATA_RANGE_TO = '~';
	
	/**
	 * The semicolon character.
	 * 
	 * This character may be used as an alternative field separator.
	 */
	private final static Character SEMICOLON = ';';

	////////////////////
	// public constants
	
	/**
	 * The set of default field separators.
	 * 
	 * This field contains only the comma character (',').
	 * It is provided as a public constant for convenience.
	 * 
	 * @since 2011-05-23
	 */
	public final static Set<Character> DEFAULT_FIELD_SEPARATOR_SET;
	
	/**
	 * The extended set of field separators.
	 * 
	 * This field contains the comma (','), the semicolon (';') and the tabulator ('\t') character.
	 * It is provided as a public constant for convenience.
	 * 
	 * @since 2011-05-23
	 */
	public final static Set<Character> EXTENDED_FIELD_SEPARATOR_SET;
	
	/**
	 * The set of characters that must never be field separators.
	 * 
	 * This field contains the carriage return ('\r'), line feed ('\n'), blank space (' ') and quotation mark ('"') characters.
	 * It is provided as a public constant for convenience.
	 * 
	 * @since 2011-05-24
	 */
	public final static Set<Character> UNSUITABLE_FIELD_SEPARATOR_SET;
	
	////////////////////
	// private members
	
	/**
	 * A buffer containing the characters that make up the current token.
	 */
	private StringBuilder buffer;
	
	/**
	 * The name of the file being scanned.
	 * 
	 * This information is required in error messages only.
	 */
	private String fileName;
	
	/**
	 * Indicates whether parsing should strictly adhere to RFC 4180 or not.
	 * TRUE  = Adhere to RFC 4180,
	 * FALSE = Be lenient.
	 */
	private boolean isStrict = true;
	
	/**
	 * The current line number.
	 */
	private int lineNumber;
	
	/**
	 * A list of the tokens produced by the parser.
	 */
	private LinkedList<Token> tokens;
	
	/**
	 * The set of characters that may be used as separators.
	 * 
	 * Usually, this set includes only the comma (',') if the parser is set to strictly adhere to RFC 4180,
	 * or the comma (','), the semicolon (';') and the tabulator character ('\t') otherwise.
	 * This set must not include double quotation marks ('"'), the carriage-return ('\r') or the line-feed character ('\n'),
	 * as these serve other purposes.
	 */
	private Set<Character> separators;
	
	/**
	 * The internal state of the scanner.
	 */
	private ScannerState state;
	
	////////////////////
	// constructors
	
	static {
		// The initialization of the constant sets has to be performed in this static block,
		// since it is too complex to be handled in one line.
		HashSet<Character> defaultSet = new HashSet<Character>();
		defaultSet.add(COMMA);
		DEFAULT_FIELD_SEPARATOR_SET = Collections.unmodifiableSet(defaultSet);
		// DEFAULT_FIELD_SEPARATOR_SET now is an unmodifiable view of the default set
		// the set itself is referenced only by variables local to this block
		// and therefore cannot be accessed from outside this block
		HashSet<Character> extendedSet = new HashSet<Character>();
		extendedSet.add(COMMA);
		extendedSet.add(SEMICOLON);
		extendedSet.add(HORIZONTAL_TABULATOR);
		EXTENDED_FIELD_SEPARATOR_SET = Collections.unmodifiableSet(extendedSet);
		// EXTENDED_FIELD_SEPARATOR_SET now is an unmodifiable view of the extended set
		// the set itself is referenced only by variables local to this block
		// and therefore cannot be accessed from outside this block
		HashSet<Character> unsuitableSet = new HashSet<Character>();
		unsuitableSet.add(CARRIAGE_RETURN);
		unsuitableSet.add(LINE_FEED);
		unsuitableSet.add(QUOTATION_MARK);
		UNSUITABLE_FIELD_SEPARATOR_SET = Collections.unmodifiableSet(unsuitableSet);
		// UNSUITABLE_FIELD_SEPARATOR_SET now is an unmodifiable view of the unsuitable set
		// the set itself is referenced only by variables local to this block
		// and therefore cannot be accessed from outside this block		
	}
	
	/**
	 * Create a new scanner!
	 * 
	 * @since 2011-09-02
	 */
	public CSVScanner() {
		this.buffer = new StringBuilder();
		this.isStrict = true;
		this.separators = new HashSet<Character>();
		this.separators.addAll(DEFAULT_FIELD_SEPARATOR_SET);
		this.state = ScannerState.END_OF_FILE;
		this.tokens = new LinkedList<Token>();
	}
	
	////////////////////
	// public methods
	
	/**
	 * Advance the state of the scanner based on the next character!
	 * 
	 * @param nextCharacter The next character.
	 * @return TRUE if at least one token was produced, FALSE otherwise.
	 * @throws IllegalCharacterException If a character is encountered that cannot be accepted.
	 * 
	 * @see #reset()
	 * @see #terminate()
	 * 
	 * @since 2011-09-02
	 */
	public boolean advanceState(char nextCharacter) throws IllegalCharacterException {
		CharacterKind kind = this.classifyCharacter(nextCharacter);
		int previousTokenCount = this.tokens.size();
		
		if (kind == CharacterKind.UNKNOWN) throw new IllegalCharacterException(this.fileName, this.lineNumber, nextCharacter);
		if (this.state == ScannerState.END_OF_FILE) throw new IllegalStateException("Input had been terminated!");
		
		switch (this.state) {
		case INITIAL_STATE:
			this.buffer.append(nextCharacter); // append the next character to the buffer in any case
			switch(kind) {
			case FIELD_SEPARATOR:                      this.createToken(TokenKind.FIELD_SEPARATOR); break; // stay in this state
			case CARRIAGE_RETURN: if (this.isStrict) { this.state = ScannerState.CARRIAGE_RETURN; break; }
			                      else               { this.state = ScannerState.LINE_BREAK;      break; }
			case LINE_FEED      : if (this.isStrict) { this.createToken(TokenKind.LINE_FEED);       break; } // stay in this state
			                      else               { this.state = ScannerState.LINE_BREAK;      break; }
			case QUOTATION_MARK :                      this.state = ScannerState.QUOTATION_MARK; break;
			case WHITESPACE     :                      this.state = ScannerState.WHITESPACE;     break;
			case TEXT           :                      this.state = ScannerState.TEXT;           break;
			}
			break;
		case CARRIAGE_RETURN: // the previous character was a carriage return; this state is reachable in strict mode only
			switch(kind) {
			case LINE_FEED      : this.buffer.append(nextCharacter); this.createToken(TokenKind.LINE_BREAK);      this.state = ScannerState.INITIAL_STATE;                                  break;
			default             :                                    this.createToken(TokenKind.CARRIAGE_RETURN); this.state = ScannerState.INITIAL_STATE; this.advanceState(nextCharacter); break; // process the next character
			}
			break;
		case LINE_BREAK: // the previous characters were any combination of carriage returns and line feeds; this state is not reachable in strict mode
			switch(kind) {
			case CARRIAGE_RETURN: // fall through
			case LINE_FEED      : this.buffer.append(nextCharacter);                                                                                              break;
			default             :                                    this.leaveLine(); this.state = ScannerState.INITIAL_STATE; this.advanceState(nextCharacter); break; // process the next character
			}
			break;
		case QUOTATION_MARK: // the previous character was a quotation mark
			switch(kind) {
			case QUOTATION_MARK : this.buffer.append(nextCharacter); this.createToken(TokenKind.ESCAPE_SEQUENCE); this.state = ScannerState.INITIAL_STATE;                                   break;
			default             :                                    this.createToken(TokenKind.QUOTATION_MARK);  this.state = ScannerState.INITIAL_STATE; this.advanceState(nextCharacter); break; // process the next character
			}
			break;
		case WHITESPACE: // the previous characters were whitespace; this state is not reachable in strict mode
			switch(kind) {
			case WHITESPACE     : this.buffer.append(nextCharacter);                                                                                                                    break; // stay in this state
			default             :                                    this.createToken(TokenKind.WHITESPACE); this.state = ScannerState.INITIAL_STATE; this.advanceState(nextCharacter); break; // process the next character
			}
			break;
		case TEXT: // the previous characters were non-whitespace
			switch(kind) {
			case TEXT           : this.buffer.append(nextCharacter);                                                                                                                   break;
			default             :                                    this.createToken(TokenKind.TEXT_DATA); this.state = ScannerState.INITIAL_STATE; this.advanceState(nextCharacter); break; // process the next character
			}
			break;
		}
		
		int newTokenCount = this.tokens.size();
		return (newTokenCount > previousTokenCount); // return TRUE if new tokens were produced
	}
	
	/**
	 * Reset this scanner to its initial state and begin scanning!
	 * 
	 * @param fileName The name of the file to be scanned. This name will only be used in error messages.
	 * 
	 * @see #advance(char)
	 * @see #terminate()
	 * 
	 * @since 2011-09-02
	 */
	public void begin(String fileName) {
		this.buffer.setLength(0);
		this.fileName = fileName;
		this.lineNumber = 1;
		this.state = ScannerState.INITIAL_STATE;
		this.tokens.clear();
	}
	
	/**
	 * Consume the oldest token that has not been consumed yet!
	 * 
	 * @return The token, or NULL if there are no unconsumed tokens.
	 * 
	 * @see #hasToken()
	 * 
	 * @since 2011-09-09
	 */
	public Token consumeToken() { return this.tokens.pollFirst(); }
	
	/**
	 * Check whether the scanner has produced tokens that have not been consumed yet!
	 * 
	 * @return TRUE if there are tokens that have not yet been consumed, FALSE otherwise.
	 * 
	 * @see #consumeToken()
	 * 
	 * @since 2011-09-09
	 */
	public boolean hasToken() { return !this.tokens.isEmpty(); }
	
	/**
	 * Define the set of acceptable field separators!
	 * 
	 * @param separators A set of characters to be accepted as field separators.
	 * @throws IllegalArgumentException If the desired set contains characters that are already used otherwise (e. g. as line breaks).
	 * 
	 * @since 2011-05-23
	 */
	public void setAcceptableFieldSeparators(Set<Character> separators) throws IllegalArgumentException {
		for (Character aCharacter : UNSUITABLE_FIELD_SEPARATOR_SET)
			if (separators.contains(aCharacter))
				throw new IllegalArgumentException(String.format("Unsuitable field separator '%c'", aCharacter));
		
		this.separators.clear();
		this.separators.addAll(separators);
	}
	
	/**
	 * Terminate the input!
	 * 
	 * @return TRUE if at least one token was produced, FALSE otherwise.
	 * 
	 * @see #reset()
	 * @see #advance(char)
	 * 
	 * @since 2011-09-12
	 */
	public boolean terminate() {
		int previousTokenCount = this.tokens.size();
		
		switch (this.state) {
		case CARRIAGE_RETURN: // the previous character was a carriage return
			this.createToken(TokenKind.CARRIAGE_RETURN); break;
		case LINE_BREAK: // the previous character was a carriage return or line feed
			this.leaveLine();                            break;
		case QUOTATION_MARK: // the previous character was a quotation mark
			this.createToken(TokenKind.QUOTATION_MARK);  break;
		case WHITESPACE: // the previous character(s) were whitespace
			this.createToken(TokenKind.WHITESPACE);      break;
		case TEXT: // the previous character(s) were non-whitespace
			this.createToken(TokenKind.TEXT_DATA);       break;
		default:
			// do nothing
		}
		
		this.state = ScannerState.END_OF_FILE;
		int newTokenCount = this.tokens.size();
		return (newTokenCount > previousTokenCount); // return TRUE if new tokens were produced
	}
	
	/**
	 * Turn strict parsing off, i. e. enable lenient parsing!
	 * 
	 * Lenient parsing means trying to continue parsing even if a violation of RFC 4180 is being detected, such as
	 * - line breaks marked by CR only, LF only or LF+CR instead of CR+LF
	 * - un-escaped quotation marks inside escaped fields.
	 * 
	 * @since 2011-05-23
	 */
	public void turnStrictScanningOff() { this.isStrict = false; }
	
	/**
	 * Turn strict parsing on, i. e. strictly adhere to RFC 4180!
	 * 
	 * Strict parsing means throwing an exception every time a violation of RFC 4180 is being detected, such as
	 * - line breaks marked by CR only, LF only or LF+CR instead of CR+LF
	 * - un-escaped quotation marks inside escaped fields.
	 * 
	 * @since 2011-05-23 
	 */
	public void turnStrictScanningOn() { this.isStrict = true; }
	
	////////////////////
	// private methods
	
	/**
	 * Classify a character!
	 * 
	 * @param aCharacter The character.
	 * @return The character kind.
	 */
	private CharacterKind classifyCharacter(char aCharacter) {
		if (this.separators.contains(aCharacter)) return CharacterKind.FIELD_SEPARATOR;
		if (aCharacter == CARRIAGE_RETURN) return CharacterKind.CARRIAGE_RETURN;
		if (aCharacter == LINE_FEED) return CharacterKind.LINE_FEED;
		if (aCharacter == QUOTATION_MARK) return CharacterKind.QUOTATION_MARK;
		if (Character.isWhitespace(aCharacter)) {
			// whitespace is treated as TEXTDATA in strict mode
			if (!this.isStrict) return CharacterKind.WHITESPACE; 
		}
		if ((aCharacter >= RFC_4180_TEXT_DATA_RANGE_FROM) && (aCharacter <= RFC_4180_TEXT_DATA_RANGE_TO)) {
			// in strict mode, characters outside this range are not allowed!
			return CharacterKind.TEXT;
		}
		if (this.isStrict) return CharacterKind.UNKNOWN;
		return CharacterKind.TEXT;
	}
	
	/**
	 * Create a new token with the given token kind at the current location!
	 * 
	 * This method empties the buffer after the token has been created.
	 * 
	 * @param kind The kind of token to be created.
	 */
	private void createToken(TokenKind kind) {
		Token tok = new Token(this.fileName, this.lineNumber, kind, this.buffer.toString());
		this.tokens.add(tok);
		this.buffer.setLength(0); // clear the buffer
	}
	
	/**
	 * Leave the current line!
	 * 
	 * This method creates a new line-break token at the current location,
	 * empties the buffer and increments the line number.
	 */
	private void leaveLine() {
		this.createToken(TokenKind.LINE_BREAK);
		++this.lineNumber;
	}
	
	////////////////////
	// local sub-classes
	
	/**
	 * An enumeration of character classes.
	 * 
	 * @since 2011-09-12
	 */
	private enum CharacterKind {
		FIELD_SEPARATOR, // a field separator
		CARRIAGE_RETURN, // the carriage return character ('\r')
		LINE_FEED,       // the line feed character ('\n')
		QUOTATION_MARK,  // double quotation marks ('"')
		WHITESPACE,      // whitespace characters
		TEXT,            // any other acceptable character
		UNKNOWN          // any unacceptable character
	}
	
	/**
     * An enumeration of the scanner states.
     * 
     * @since 2011-09-02
     */
    private enum ScannerState {
    	INITIAL_STATE,   // initial state
    	CARRIAGE_RETURN, // a carriage return has been read and the scanner needs to check the next character
    	LINE_BREAK,      // a sequence of carriage returns and line feeds has been read and the scanner expects more
    	QUOTATION_MARK,  // a quotation mark has been read and the scanner needs to check the next character
    	WHITESPACE,      // whitespace has been read and the scanner expects more
    	TEXT,            // non-whitespace characters have been read and the scanner expects more
    	END_OF_FILE      // end of file
    }

}
