/*
 Copyright (c) 2007--2012 Technische Universitt Darmstadt.
 
 Permission is hereby granted to use, copy, modify, and distribute this
 software, provided that this copyright notice remains intact.
 */
package ptolemy.actor.lib.io;

import gsi.rf.data.InvalidCharacterException;
import gsi.rf.data.ParseException;
import gsi.rf.data.csv.CSVDocument;
import gsi.rf.data.csv.CSVReader;
import gsi.rf.data.csv.CSVRecord;
import java.io.IOException;
import ptolemy.actor.TypedIOPort;
import ptolemy.actor.lib.Source;
import ptolemy.data.BooleanToken;
import ptolemy.data.IntToken;
import ptolemy.data.StringToken;
import ptolemy.data.expr.FileParameter;
import ptolemy.data.expr.Parameter;
import ptolemy.data.type.BaseType;
import ptolemy.kernel.CompositeEntity;
import ptolemy.kernel.util.Attribute;
import ptolemy.kernel.util.IllegalActionException;
import ptolemy.kernel.util.NameDuplicationException;
import ptolemy.kernel.util.Workspace;


//////////////////////////////////////////////////////////////////////////
//// CSVRecordReader

/**
   This actor reads a file or URL in CSV (character-separated values) format,
   one line at a time, and outputs each field within the line on a different
   channel of its output multiport. The file or URL is specified using any
   form acceptable to FileParameter.
   Before an end of file is reached, the <i>endOfFile</i> output produces
   <i>false</i>.  In the iteration where the last line of the file is read and
   produced on the <i>output</i> port, this actor produces <i>true</i> on the
   <i>endOfFile</i> port. In that iteration, postfire() returns false.  If the
   actor is iterated again, after the end of file, then prefire() and postfire()
   will both return false, <i>output</i> will produce an empty array, and
   <i>endOfFile</i> will produce <i>true</i>.
   <p>
   In some domains (such as SDF), returning false in postfire()
   causes the model to cease executing.
   In other domains (such as DE), this causes the director to avoid
   further firings of this actor.  So usually, the actor will not be
   invoked again after the end of file is reached.
   <p>
   This actor reads ahead in the file so that it can produce an output
   <i>true</i> on <i>endOfFile</i> in the same iteration where it outputs
   the last line.  It reads the first line in preinitialize(), and
   subsequently reads a new line in each invocation of postfire().  The
   line read is produced on the <i>output</i> in the next iteration
   after it is read.
   <p>
   This actor can skip some lines at the beginning of the file or URL, with
   the number specified by the <i>numberOfLinesToSkip</i> parameter. The
   default value of this parameter is 0.
   <p>
   If one needs to reset this CSV reader to start again at the beginning
   of the file, the way to do this is to call initialize() during the run
   of the model.  This can be done, for example, using a modal model
   with a transition where reset is enabled.

   @see LineReader
   @author Christopher Spies
*/
public class CSVRecordReader extends Source {
    /** Construct an actor with the given container and name.
     *  @param container The container.
     *  @param name The name of this actor.
     *  @exception IllegalActionException If the actor cannot be contained
     *   by the proposed container.
     *  @exception NameDuplicationException If the container already has an
     *   actor with this name.
     */
    public CSVRecordReader(CompositeEntity container, String name)
            throws IllegalActionException, NameDuplicationException {
        super(container, name);

	output.setMultiport(true);
        output.setTypeEquals(BaseType.STRING);

        endOfFile = new TypedIOPort(this, "endOfFile", false, true);
        endOfFile.setTypeEquals(BaseType.BOOLEAN);

        fileOrURL = new FileParameter(this, "fileOrURL");

        numberOfLinesToSkip = new Parameter(this, "numberOfLinesToSkip");
        numberOfLinesToSkip.setExpression("0");
        numberOfLinesToSkip.setTypeEquals(BaseType.INT);
        
        _attachText("_iconDescription",
                "<svg>\n" + "<rect x=\"-25\" y=\"-20\" "
                + "width=\"50\" height=\"40\" " + "style=\"fill:white\"/>\n"
                + "<polygon points=\"-15,-10 -12,-10 -8,-14 -1,-14 3,-10"
                + " 15,-10 15,10, -15,10\" " + "style=\"fill:red\"/>\n"
                + "</svg>\n");
        
        // the reader must be initialized here, otherwise it might not have been initialized properly before _readFile() is first called
        try {
        	_reader = new CSVReader();
        	_reader.setAcceptableFieldSeparators(CSVReader.DEFAULT_FIELD_SEPARATOR_SET);
        } catch (InvalidCharacterException icx) {
        	throw new IllegalActionException(this, icx, "Initialize failed");
        }
    }

    ///////////////////////////////////////////////////////////////////
    ////                     ports and parameters                  ////

    /** An output port that produces <i>false</i> until the end of file
     *  is reached, at which point it produces <i>true</i>. The type
     *  is boolean.
     */
    public TypedIOPort endOfFile;

    /** The file name or URL from which to read.  This is a string with
     *  any form accepted by FileParameter.
     *  @see FileParameter
     */
    public FileParameter fileOrURL;

    /** The number of lines to skip at the beginning of the file or URL.
     *  This parameter contains an IntToken, initially with value 0.
     *  The value of this parameter must be non-negative.
     */
    public Parameter numberOfLinesToSkip;

    ///////////////////////////////////////////////////////////////////
    ////                         public methods                    ////

    /** If the specified attribute is <i>fileOrURL</i> and there is an
     *  open file being read, then close that file and open the new one;
     *  if the attribute is <i>numberOfLinesToSkip</i> and its value is
     *  negative, then throw an exception.  In the case of <i>fileOrURL</i>,
     *  do nothing if the file name is the same as the previous value of
     *  this attribute.
     *  @param attribute The attribute that has changed.
     *  @exception IllegalActionException If the specified attribute
     *   is <i>fileOrURL</i> and the file cannot be opened, or the previously
     *   opened file cannot be closed; or if the attribute is
     *   <i>numberOfLinesToSkip</i> and its value is negative.
     */
    public void attributeChanged(Attribute attribute)
            throws IllegalActionException {
        if (attribute == fileOrURL) {
            // NOTE: We do not want to close the file if the file
            // has not in fact changed.  We check this by just comparing
            // name, which is not perfect...
            String newFileName = ((StringToken)fileOrURL.getToken()).stringValue();

            if ((_previousFileOrURL == null)
             || !newFileName.equals(_previousFileOrURL))
            {
           		_document = null;
            	_previousFileOrURL = newFileName;
            }
        } else if (attribute == numberOfLinesToSkip) {
            int linesToSkip = ((IntToken) numberOfLinesToSkip.getToken()).intValue();

            if (linesToSkip < 0) throw new IllegalActionException(this, "The number of lines to skip cannot be negative.");
        } else {
            super.attributeChanged(attribute);
        }
    }

    /** Clone the actor into the specified workspace.
     *  @param workspace The workspace for the new object.
     *  @return A new actor.
     *  @exception CloneNotSupportedException If a derived class contains
     *   an attribute that cannot be cloned.
     */
    public Object clone(Workspace workspace) throws CloneNotSupportedException {
        CSVRecordReader newObject = (CSVRecordReader)super.clone(workspace);
        newObject._document = null;
        newObject._firedSinceWrapup = false;
        newObject._previousFileOrURL = null;
        newObject._recordIndex = 0;
        return newObject;
    }

    /** Output the data read in the preinitialize() or in the previous
     *  invocation of postfire(), if there is any.
     *  @exception IllegalActionException If there's no director.
     */
    public void fire() throws IllegalActionException {
        super.fire();

        if (!_isEndOfFileReached()) {
        	CSVRecord record = _document.getRecord(_recordIndex);
            for (int fieldIndex=0; fieldIndex<record.countFields(); ++fieldIndex) output.send(fieldIndex, new StringToken(record.getFieldContent(fieldIndex)));
        }
    }

    /** If this is called after prefire() has been called but before
     *  wrapup() has been called, then close any
     *  open file, re-open it, skip the number of lines given by the
     *  <i>numberOfLinesToSkip</i> parameter, and read the first line to
     *  be produced in the next invocation of prefire(). This occurs if
     *  this actor is re-initialized during a run of the model.
     *  @exception IllegalActionException If any error occurs while trying to open and read the file.
     */
    public void initialize() throws  IllegalActionException {
        super.initialize();
        
        // It would be better if there were a way to reset the
        // input stream, but apparently there is not, short of
        // rereading the file.
        if (_firedSinceWrapup) _document = null;
        if (_document == null) _readFile();
    }

    /** Read the next line from the file. If there is no next line,
     *  return false.  Otherwise, return whatever the superclass returns.
     *  @exception IllegalActionException If there is a problem reading
     *   the file.
     */
    public boolean postfire() throws IllegalActionException {
   		_recordIndex ++; // go to the next record

        if (_isEndOfFileReached()) {
            endOfFile.broadcast(BooleanToken.TRUE);
            return false;
        } else {
            endOfFile.broadcast(BooleanToken.FALSE);
            return super.postfire();
        }
    }

    /** Return false if there is no more data available in the file.
     *  Otherwise, return whatever the superclass returns.
     *  @exception IllegalActionException If the superclass throws it.
     */
    public boolean prefire() throws IllegalActionException {
        _firedSinceWrapup = true;
        
        if (_isEndOfFileReached()) return false;
        return super.prefire();
    }

    /** 
     * Open the file or URL,
     * skip the number of lines specified by the <i>numberOfLinesToSkip</i> parameter,
     * and read the file's content.
     *  
     * @throws IllegalActionException If any error occurs while trying to open and read the file.
     */
    public void preinitialize() throws IllegalActionException {
        super.preinitialize();
        _readFile();
        _firedSinceWrapup = false;
    }

    /** Close the reader if there is one.
     *  @exception IllegalActionException If an IO error occurs.
     */
    public void wrapup() throws IllegalActionException {
        _firedSinceWrapup = false;
    }

    /////////////////////////////////////////////////////////////////////
    ////                         protected methods                   ////

    /**
     * Check whether EOF has been reached!
     * 
     * @return TRUE if the last record has been read
     *         FALSE otherwise.
     */
    protected boolean _isEndOfFileReached() {
        return (_recordIndex >= _document.countRecords());
    }

    ///////////////////////////////////////////////////////////////////
    ////                         private methods                   ////
    
    /**
     * [Re]Open the file and skip as many lines as given by the {@link #numberOfLinesToSkip} parameter!
     * 
     * @throws IllegalActionException If the file cannot be opened or contains less lines than should be skipped.
     * @throws IOException If any error occurs while reading the file.
     * @throws ParseException If the file cannot be parsed.
     */
    private void _readFile() throws IllegalActionException {
    	try {
    		_document = _reader.read(fileOrURL.asFile());
    	} catch (IOException iox) {
    		throw new IllegalActionException(this, iox, "Error reading file");
    	} catch (ParseException px) {
    		throw new IllegalActionException(this, px, "Error reading file");
    	}
        _recordIndex = ((IntToken)numberOfLinesToSkip.getToken()).intValue(); // skip the first lines
    }
    
    ///////////////////////////////////////////////////////////////////
    ////                         private members                   ////
    
    /**
     * The CSV document being read.
     */
    private CSVDocument _document = null;
    
    /** Indicator that the fire() method has been called, but wrapup
     *  has not.  That is, we are in the middle of a run.
     */
    private boolean _firedSinceWrapup = false;
    
    /**
     * The index of the current record inside the document.
     */
    private int _recordIndex = 0;

    /**
     * Previous value of fileOrURL parameter.
     * This is required in order to be able to determine whether the file has changed.
     */
    private String _previousFileOrURL = null;
    
    /**
     * The CSV reader being used by this class.
     */
    private CSVReader _reader = null;
    
}
