/*
 Copyright (c) 2007--2012 Technische Universitt Darmstadt.
 
 Permission is hereby granted to use, copy, modify, and distribute this
 software, provided that this copyright notice remains intact.
 */
package ptolemy.actor.lib.io;

import gsi.rf.data.csv.CSVDocument;
import gsi.rf.data.csv.CSVField;
import gsi.rf.data.csv.CSVRecord;
import gsi.rf.data.csv.CSVWriter;

import java.io.File;
import java.io.IOException;

import ptolemy.actor.lib.Sink;
import ptolemy.data.ArrayToken;
import ptolemy.data.BooleanToken;
import ptolemy.data.StringToken;
import ptolemy.data.expr.FileParameter;
import ptolemy.data.expr.Parameter;
import ptolemy.data.type.ArrayType;
import ptolemy.data.type.BaseType;
import ptolemy.kernel.CompositeEntity;
import ptolemy.kernel.util.Attribute;
import ptolemy.kernel.util.IllegalActionException;
import ptolemy.kernel.util.NameDuplicationException;
import ptolemy.kernel.util.Workspace;
import ptolemy.util.MessageHandler;

//////////////////////////////////////////////////////////////////////////
////CSVRecordReader

/**
   <p>This actor reads input tokens and writes them to a file or URL in CSV (character-
   separated values) format. Each channel of the input multiport becomes a different
   field in the resulting CSV record. The file or URL is specified using any
   form acceptable to FileParameter.</p>
   <p>
   If the <i>append</i> attribute has value <i>true</i>,
   then the file will be appended to. If it has value <i>false</i>,
   then if the file exists, the user will be queried for permission
   to overwrite, and if granted, the file will be overwritten.</p>
   <p>
   If the <i>confirmOverwrite</i> parameter has value <i>false</i>,
   then this actor will overwrite the specified file if it exists
   without asking.  If <i>true</i> (the default), then if the file
   exists, then this actor will ask for confirmation before overwriting.</p>

   @see LineReader
   @author Christopher Spies
*/

public class CSVRecordWriter extends Sink {
	
	/**
	 * The writer buffers this many fields before it flushes them to the output file.
	 */
	private static final int BUFFER_SIZE = 1048576;

	/**
	 * The Serial Version UID of this class, required by all serializable classes.
	 */
	private static final long serialVersionUID = -3567301964549503293L;

	/** Construct an actor with the given container and name.
     *  @param container The container.
     *  @param name The name of this actor.
     *  @exception IllegalActionException If the actor cannot be contained
     *   by the proposed container.
     *  @exception NameDuplicationException If the container already has an
     *   actor with this name.
     */
    public CSVRecordWriter(CompositeEntity container, String name)
            throws IllegalActionException, NameDuplicationException {
        super(container, name);

        input.setMultiport(true);
        input.setTypeEquals(BaseType.UNKNOWN);

        fileOrURL = new FileParameter(this, "fileOrURL");
        
        headlines = new Parameter(this, "headlines");
        headlines.setTypeEquals(new ArrayType(BaseType.STRING));
        headlines.setExpression("{ \"t / s\" }");

        append = new Parameter(this, "append");
        append.setTypeEquals(BaseType.BOOLEAN);
        append.setToken(BooleanToken.FALSE);

        confirmOverwrite = new Parameter(this, "confirmOverwrite");
        confirmOverwrite.setTypeEquals(BaseType.BOOLEAN);
        confirmOverwrite.setToken(BooleanToken.TRUE);
        
        removeDuplicates = new Parameter(this, "removeDuplicates");
        removeDuplicates.setTypeEquals(BaseType.BOOLEAN);
        removeDuplicates.setToken(BooleanToken.TRUE);

        _attachText("_iconDescription", "<svg>\n"
                + "<rect x=\"-25\" y=\"-20\" " + "width=\"50\" height=\"40\" "
                + "style=\"fill:white\"/>\n"
                + "<polygon points=\"-15,-10 -12,-10 -8,-14 -1,-14 3,-10"
                + " 15,-10 15,10, -15,10\" " + "style=\"fill:red\"/>\n"
                + "</svg>\n");
        
      	_writer = new CSVWriter();
       	_writer.turnTrimmingOn();
    }
    
    ///////////////////////////////////////////////////////////////////
    ////                     ports and parameters                  ////

    /** If <i>true</i>, then append to the specified file.  If <i>false</i>
     *  (the default), then overwrite any preexisting file after asking
     *  the user for permission.
     */
    public Parameter append;
    
    /** If <i>false</i>, then overwrite the specified file if it exists
     *  without asking.  If <i>true</i> (the default), then if the file
     *  exists, ask for confirmation before overwriting.
     */
    public Parameter confirmOverwrite;
    
    /** If <i>false</i>, then write data as is.
     *  If <i>true</i> (the default), then remove duplicate
     *  records before writing the file.
     */
    public Parameter removeDuplicates;

    /** The file name to which to write.  This is a string with
     *  any form accepted by FileParameter.  The default value is
     *  "System.out".
     *  @see FileParameter
     */
    public FileParameter fileOrURL;

    /** Headline to print at the beginning of the file.
     *  This parameter contains a StringToken.
     */
    public Parameter headlines;
    
    ///////////////////////////////////////////////////////////////////
    ////                         public methods                    ////

    /** If the specified attribute is <i>fileName</i> and there is an
     *  open file being written, then close that file.  The new file will
     *  be opened or created when it is next written to.
     *  @param attribute The attribute that has changed.
     *  @exception IllegalActionException If the specified attribute
     *   is <i>fileName</i> and the previously
     *   opened file cannot be closed.
     */
    public void attributeChanged(Attribute attribute)
            throws IllegalActionException {
        if (attribute == fileOrURL) {
            // Do not change the file if it has the same name
            String newFileName = ((StringToken)fileOrURL.getToken()).stringValue();

            if ((_previousFileOrURL == null)
            || !newFileName.equals(_previousFileOrURL))
            {
               	_previousFileOrURL = newFileName;
               	_isWriteAuthorized = false;
            }
        } else {
            super.attributeChanged(attribute);
        }
    }
    
    /** Clone the actor into the specified workspace.
     *  @param workspace The workspace for the new object.
     *  @return A new actor.
     *  @exception CloneNotSupportedException If a derived class contains
     *   an attribute that cannot be cloned.
     */
    public Object clone(Workspace workspace) throws CloneNotSupportedException {
        CSVRecordWriter newObject = (CSVRecordWriter)super.clone(workspace);
        newObject._previousFileOrURL = null;
        newObject._isWriteAuthorized = false;
        return newObject;
    }
    
    /**
     * Check whether the specified file already exists,
     * and prompt the user for confirmation if so.
     * 
     * @throws IllegalActionException If the user selects not to overwrite the file.
     */
    public void initialize() throws  IllegalActionException {
        super.initialize();
        
        if (!_isWriteAuthorized) _getUserApproval();
        if (_document == null) _document = new CSVDocument(); else _document.clear();
        
        // create the headline:
        ArrayToken headers = (ArrayToken)headlines.getToken();
        CSVRecord headlineRecord = new CSVRecord();
        for (int fieldIndex=0; fieldIndex<headers.length(); ++fieldIndex) {
        	StringToken aHeader = (StringToken)headers.getElement(fieldIndex);
        	headlineRecord.addField(new CSVField(aHeader.stringValue()));
        }
        _document.addRecord(headlineRecord);
        
        // create the output file:
        _currentFile = fileOrURL.asFile();
        _beginWrite();
    }
    
    /** Read input string tokens and write them to the output cache.
     *  If there is no input, do nothing.
     */
    public boolean postfire() throws IllegalActionException {
    	// create the next line:
    	int channelCount = input.getWidth();
        CSVRecord aRecord = new CSVRecord();
        for (int channelIndex=0; channelIndex<channelCount; ++channelIndex)
        	aRecord.addField(new CSVField(input.get(channelIndex).toString()));
        _document.addRecord(aRecord);
        
        // write the buffered data to the output file if necessary:
        if (_document.countRecords() * channelCount > BUFFER_SIZE) this._flush();
    	
        return super.postfire();
    }
    
    /** Close the writer if there is one.
     *  @exception IllegalActionException If an IO error occurs.
     */
    public void wrapup() throws IllegalActionException {
        this._flush();

        _isWriteAuthorized = false;
    }
    
    ///////////////////////////////////////////////////////////////////
    ////                         private methods                   ////
    
    /**
     * Begin writing to the specified file!
     * 
     * @throws IllegalActionException If an error occurs accessing the file.
     */
    private void _beginWrite() throws IllegalActionException {
    	BooleanToken appendToken = (BooleanToken)append.getToken();
    	boolean appendValue      = appendToken.booleanValue();
    	
    	System.out.println(String.format("Opening '%s'...", _currentFile.getName()));
        this._writeDocument(appendValue);
    }
    
    /**
     * Flush buffered results to the specified file!
     * 
     * @throws IllegalActionException If an error occurs accessing the file.
     */
    private void _flush() throws IllegalActionException {
    	System.out.println(String.format("Saving results to '%s'...", _currentFile.getName()));
    	this._writeDocument(true);
    }
    
    /**
     * Get the user's approval for writing to the specified file!
     * 
     * Writing to System.out is always considered OK.
     * If the file does not yet exist, it is always assumed OK to create it.
     * If the file is only appended to, the user's approval is implicitly assumed.
     * If the 'confirm overwrite' parameter is set to FALSE,
     * the user's approval to overwriting existing files is also implicitly assumed.
     * Only if none of the above conditions apply is the user queried for his approval to overwrite an existing file.
     * 
     * @throws IllegalActionException If an error occurs accessing the file
     *                                or the user answers 'No' when queried about overwriting an existing file.
     */
    private void _getUserApproval() throws IllegalActionException {
        BooleanToken appendToken           = (BooleanToken) append.getToken();
        BooleanToken confirmOverwriteToken = (BooleanToken) confirmOverwrite.getToken();
        boolean appendValue           = appendToken.booleanValue();
        boolean confirmOverwriteValue = confirmOverwriteToken.booleanValue();

        if (fileOrURL.stringValue().equals("System.out")) {
            // Only check for append and overwrite if the
            // fileName is not "System.out"
            _isWriteAuthorized = true;
        } else {
            // then open the file.
            File file = fileOrURL.asFile();

            // Don't ask for confirmation in append mode, since there
            // will be no loss of data.
            if (file.exists()) {
                if (appendValue) {
                    _isWriteAuthorized = true;
                } else {
                    if (confirmOverwriteValue) {
                        // Query for overwrite.
                        // FIXME: This should be called in the event thread!
                        // There is a chance of deadlock since it is not.
                        if (MessageHandler.yesNoQuestion("OK to overwrite " + file + "?")) {
                            _isWriteAuthorized = true;
                        } else {
                            throw new IllegalActionException(this, "Please select another file name.");
                        }
                    } else { // overwrite does not need to be confirmed
                        _isWriteAuthorized = true;
                    }
                }
            } else { // file does not exist (yet)
                _isWriteAuthorized = true;
            }
        }
    }
    
    /**
     * Write the current state of the document to the specified file,
     * then clear the document!
     * 
     * @param append TRUE if the document is to be appended to the specified file,
     *               FALSE if the specified file is to be overwritten.
     * @throws IllegalActionException If an error occurs accessing the file.
     */
    private void _writeDocument(boolean append) throws IllegalActionException {
    	BooleanToken removeDuplicatesToken = (BooleanToken)removeDuplicates.getToken();
    	boolean removeDuplicatesValue      = removeDuplicatesToken.booleanValue();
    	
    	_document.trim();
        if (removeDuplicatesValue) _document.removeDuplicateRecords();
    	try {
            _writer.write(_document, _currentFile, append);
            _document.clear();
        } catch (IOException ioex) {
            throw new IllegalActionException(this, ioex, String.format("Error writing to '%s': %s!", _currentFile.getName(), ioex.getMessage()));
        }
    }
    
    ///////////////////////////////////////////////////////////////////
    ////                         private members                   ////
    
    /**
     * The file to be written to.
     * This is required in case the fileOrURL parameter is changed during simulation.
     */
    private File _currentFile;
    
    /**
     * The CSV document being written.
     */
    private CSVDocument _document = null;
    
    /**
     * Previous value of fileOrURL parameter.
     * This is required in order to be able to determine whether the file has changed.
     */
    private String _previousFileOrURL = null;
    
    /**
     * Indicator that the user has authorized to overwrite the file (if this is necessary and required).
     */
    private boolean _isWriteAuthorized = false;
    
    /**
     * The CSV writer being used by this class.
     */
    private CSVWriter _writer = null;
    
}
