/*
 Copyright (c) 2010 Technische Universitt Darmstadt.
 
 Permission is hereby granted to use, copy, modify, and distribute this
 software, provided that this copyright notice remains intact.
 */

package ptolemy.batch;

import java.util.List;

import tcl.lang.Command;
import tcl.lang.Interp;
import tcl.lang.TclException;
import tcl.lang.TclInteger;
import tcl.lang.TclList;
import tcl.lang.TclNumArgsException;
import tcl.lang.TclObject;
import tcl.lang.TclString;

import ptolemy.actor.CompositeActor;
import ptolemy.actor.TypedIOPort;
import ptolemy.actor.TypedIORelation;
import ptolemy.kernel.ComponentEntity;
import ptolemy.kernel.CompositeEntity;
import ptolemy.kernel.util.IllegalActionException;
import ptolemy.kernel.util.NameDuplicationException;

/**
 * This class implements a new Tcl command "ptgenerate"
 * that replicates an actor within a Ptolemy II model a given number of times.
 * 
 * @author cspies
 */
public class PtGenerateCommand implements Command {

    /**
     * Invoked by the interpreter to handle the "ptgenerate" command.
     * 
     * The command syntax is:
     * 
     *     ptgenerate handle name number
     * 
     * The Ptolemy II model specified by 'handle'
     * is searched for the actor specified by 'name'.
     * This actor is then replicated 'number' times.
     * The original actor is renamed ('_0' is appended to its name),
     * while its clones are given the name of the original actor plus '_1', '_2' etc.
     * A list containing the names of the generated instances is returned.
     * If the handle is invalid, the actor does not exist or the replication fails,
     * an error message is returned as the Tcl result object.
     * 
     * @param interp The interpreter used to execute the command.
     * @param objv The parameters of the model.
     *             Besides the command name, three parameters are expected.
     */
    public void cmdProc(Interp interp, TclObject[] objv) throws TclException {
        // local variables:
        String[] cloneNames;
        TclObject result;
        CompositeActor toplevel;
        ComponentEntity original;
        String originalName;
        int numberOfInstances;
        
        // get input:
        if (objv.length != 4) throw new TclNumArgsException(interp, 1, objv, "actorHandle actorName instanceCount"); // make sure there is exactly one command and three parameters
        toplevel = TclPtolemyActor.get(interp, objv[1]);
        originalName = objv[2].toString();
        numberOfInstances = TclInteger.get(interp, objv[3]);
        
        // perform action:
        original = toplevel.getEntity(originalName);
        if (original == null) throw new TclException(interp, String.format("No actor named '%s' exists inside actor '%s'!", originalName, toplevel.getName()));
        try {
            cloneNames = replicateActor(original, numberOfInstances);
        } catch (CloneNotSupportedException cnsex) {
            throw new TclException(interp, String.format("Error replicating actor '%s': %s", original.getName(), cnsex.getMessage()));
        } catch (IllegalActionException iaex) {
            throw new TclException(interp, String.format("Error replicating actor '%s': %s", original.getName(), iaex.getMessage()));
        } catch (NameDuplicationException ndex) {
            throw new TclException(interp, String.format("Error replicating actor '%s': %s", original.getName(), ndex.getMessage()));
        }

        // produce output:
        result = TclList.newInstance();
        for (String aCloneName : cloneNames) TclList.append(interp, result, TclString.newInstance(aCloneName));
        interp.setResult(result);
    }

    /**
     * Connects one port of a replicate to the ports that the same port on the original was connected to.
     * 
     * @param clonePort The port on the replicated actor corresponding to the port of the original actor.
     * @param clone The replicated actor that is to be connected.
     * @param originalPort The port of the original actor.
     * @throws IllegalActionException   If such an exception is thrown when trying to create a new relation
     *                                  or the original port is connected to more than one relation
     *                                  or an output port on the original actor is connected to a remote port that is not a multiport.
     * @throws NameDuplicationException If such an exception is thrown when trying to create a new relation.
     */
    private static void connectReplicatePort(TypedIOPort clonePort, ComponentEntity clone, TypedIOPort originalPort) throws IllegalActionException, NameDuplicationException {
        List portList;
        CompositeEntity container;
        TypedIORelation originalRelation, cloneRelation;
        TypedIOPort remotePort;
        
        assert clonePort != null;
        assert clone != null;
        assert originalPort != null;
        assert clone.portList().contains(clonePort);
        assert clonePort.getName().equals(originalPort.getName());
        
        container = (CompositeEntity)clone.getContainer(); // == original.getContainer()
        
        originalRelation = getRelation(originalPort);
        if (originalRelation == null) return; // no relation -> no ports -> nothing to do

        portList = getRelevantRemotePorts(originalPort, originalRelation);
        if (portList == null) return; // no ports -> nothing to do
        
        cloneRelation = originalRelation; // reuse the original relation unless the list of remote ports contains multiports
        if (containsMultiports(portList)) {
            cloneRelation = new TypedIORelation(container, String.format("r_%s_%s", clone.getName(), clonePort.getName()));
            System.out.println(String.format("%s: created relation %s", originalPort.getFullName(), cloneRelation.getFullName())); // debug
            
            for (Object anotherPort : portList) {
                remotePort = (TypedIOPort)anotherPort;
                if (clonePort.isOutput() && !remotePort.isMultiport()) throw new IllegalActionException(originalPort, "Must be connected to a multi-port");
                remotePort.link(cloneRelation);
            }
        }
        clonePort.link(cloneRelation);
        System.out.println(String.format("%s: linked to %s", clonePort.getFullName(), cloneRelation.getFullName())); // debug
    }
    
    /**
     * Determine whether a list of ports contains multiports!
     * 
     * @param portList The list of ports.
     * @return TRUE if at least one port in the list is a multiport, FALSE otherwise.
     */
    private static boolean containsMultiports(List portList) {
        TypedIOPort port;
        
        assert portList != null;
        for (Object element : portList) assert element instanceof TypedIOPort;
        
        for (Object element : portList) {
            port = (TypedIOPort)element;
            if (port.isMultiport()) return true;
        }
        return false;
    }
    
    /**
     * Get the single linked relation for a given port!
     * If the port is not linked to any relation, returns NULL.
     * If the port is linked to more than one relation, an exception is thrown.
     * 
     * @param port The port.
     * @return The single relation the port is linked to, or NULL if no such relation exists.
     * @throws IllegalActionException If the port is connected to more than one relation.
     */
    private static TypedIORelation getRelation(TypedIOPort port) throws IllegalActionException {
        List relationList;
        TypedIORelation relation;
        
        assert port != null;
        
        relationList = port.linkedRelationList();
        if (relationList == null) return null;
        if (relationList.size() < 1) return null; // no relation -> no ports -> nothing to do
        if (relationList.size() > 1) throw new IllegalActionException(port, "Can be linked to one relation only");
        relation = (TypedIORelation)relationList.get(0);
        
        return relation;
    }
    
    /**
     * Get the relevant remote ports a given port communicates with over a given relation!
     * The relevant ports are the input ports an output is connected to on the outside,
     *                        the output ports an output is connected to on the inside,
     *                        the input ports an input is connected to on the inside and
     *                        the output ports an input is connected to on the outside.
     * 
     * @param port The port.
     * @param relation The relation by which the port is linked to the remote ports.
     * @return A list containing the relevant remote ports, or NULL if no such ports exist.
     */
    private static List getRelevantRemotePorts(TypedIOPort port, TypedIORelation relation) {
        List portList;
        
        assert port != null;
        assert relation != null;
        assert port.linkedRelationList().contains(relation);
        assert relation.linkedPortList().contains(port);
        
        portList = null;
        if (port.isInput())  portList = relation.linkedSourcePortList(port);
        if (port.isOutput()) portList = relation.linkedDestinationPortList(port);
        if (portList == null) return null;
        if (portList.isEmpty()) return null;
        return portList;
    }
   
    /**
     * Replicate the given actor a given number of times and return a list with the names of the generated replicates!
     * The original's name will be changed as well.
     * 
     * @param original The actor to replicate.
     * @param instanceCount The number of times the actor shall be replicated (>= 1).
     * @return An array containing the names of the original actor as well as all the replicates, in the order they were created.
     *         The original's new name is the first element in the array.
     *         The length of the array is equal to instanceCount.
     * @throws CloneNotSupportedException If such an exception is thrown when trying to clone the original actor.
     * @throws IllegalActionException     If such an exception is thrown when trying to rename the original actor
     *                                                                or when trying to set up or rename a replicate actor
     *                                                                or when trying to create a new relation
     *                                    or if a port of the original actor is connected to more than one relation.
     *                                    or an output port on the original actor is connected to a remote port that is not a multiport.
     * @throws NameDuplicationException   If such an exception is thrown when trying to rename the original actor.
     *                                                                or when trying to create a new relation.
     */
    private static String[] replicateActor(ComponentEntity original, int instanceCount) throws CloneNotSupportedException, IllegalActionException, NameDuplicationException {
        String cloneNames[];
        String originalName;
        int currentInstance;
        
        assert original != null;
        assert instanceCount >= 1;

        cloneNames = new String[instanceCount];
        originalName = original.getName();
        cloneNames[0] = String.format("%s_%d", originalName, 0);
        for (currentInstance=1; currentInstance<instanceCount; ++currentInstance)
            cloneNames[currentInstance] = replicateActorOnce(original, currentInstance);
        original.setName(cloneNames[0]);
        
        return cloneNames;
    }
    
    /**
     * Replicate the given actor once and return the names of the generated replicate!
     * 
     * @param original The actor to replicate.
     * @param instance The instance of the replicate, starting from 1 (1 = first clone, 2 = second clone, ...).
     * @return The names of the new replicate.
     * @throws CloneNotSupportedException If such an exception is thrown when trying to clone the original actor.
     * @throws IllegalActionException     If such an exception is thrown when trying to set up or rename the replicate actor
     *                                                                or when trying to create a new relation
     *                                    or if a port of the original actor is connected to more than one relation.
     *                                    or an output port on the original actor is connected to a remote port that is not a multiport.
     * @throws NameDuplicationException   If such an exception is thrown when trying to rename the original actor
     *                                                                or when trying to create a new relation.
     */
    private static String replicateActorOnce(ComponentEntity original, int instanceIndex) throws CloneNotSupportedException, IllegalActionException, NameDuplicationException {
        CompositeEntity container;
        ComponentEntity clone;
        TypedIOPort originalPort, clonePort;
        String originalName, cloneName;

        assert original != null;
        assert instanceIndex >= 1;
        
        originalName = original.getName();
        cloneName = String.format("%s_%d", originalName, instanceIndex);
        container = (CompositeEntity)original.getContainer();
        clone = (ComponentEntity)original.clone(container.workspace());
        clone.setName(cloneName);
        clone.setContainer(container);
        clone.validateSettables();
        System.out.println(String.format("Cloned: %s", clone.getFullName())); // debug
        
        for (Object aPort : original.portList()) {
            originalPort = (TypedIOPort)aPort;
            clonePort = (TypedIOPort)clone.getPort(originalPort.getName()); // find a port with the same name in the clone
            connectReplicatePort(clonePort, clone, originalPort);
        }
        
        return cloneName;
    }

}
