/*
 Copyright (c) 2010 Technische Universitt Darmstadt.
 
 Permission is hereby granted to use, copy, modify, and distribute this
 software, provided that this copyright notice remains intact.
 */

package ptolemy.batch;

import tcl.lang.Command;
import tcl.lang.Interp;
import tcl.lang.TclException;
import tcl.lang.TclNumArgsException;
import tcl.lang.TclObject;

import ptolemy.actor.CompositeActor;
import ptolemy.data.expr.Parameter;
import ptolemy.kernel.util.IllegalActionException;

/**
 * This class implements a new Tcl command "ptgetparameter"
 * that reads a parameter within a Ptolemy II model and returns its value.
 * 
 * @author cspies
 */
public class PtGetParameterCommand implements Command {

    /**
     * Invoked by the interpreter to handle the "ptgetparameter" command.
     * 
     * The command syntax is:
     * 
     *     ptgetparameter handle name
     * 
     * The Ptolemy II model specified by 'handle'
     * is searched for the parameter specified by 'name',
     * and that parameter is read.
     * The current value of the parameter is returned.
     * If the handle is invalid or the parameter does not exist,
     * an error message is returned as the Tcl result object.
     * 
     * @param interp The interpreter used to execute the command.
     * @param objv The parameters of the model.
     *             Besides the command name, three parameters are expected.
     */
    public void cmdProc(Interp interp, TclObject[] objv) throws TclException {
        // local variables:
        CompositeActor toplevel;
        Parameter parameter;
        String parameterName;
        
        // get input:
        if (objv.length != 3) throw new TclNumArgsException(interp, 1, objv, "actorHandle parameterName"); // make sure there is exactly one command and two parameters
        toplevel = TclPtolemyActor.get(interp, objv[1]);
        parameterName = objv[2].toString();
        
        // perform action:
        parameter = (Parameter)toplevel.getAttribute(parameterName);
        if (parameter == null) throw new TclException(interp, String.format("No parameter named '%s' exists inside actor '%s'!", parameterName, toplevel.getName()));
        
        // produce output:
        try {
            interp.setResult(parameter.getToken().toString()); // return the string representation of the new value so that the calling code can check it
        } catch (IllegalActionException iaex) {
            throw new TclException(interp, String.format("Error accessing parameter '%s': %s", parameterName, iaex.toString()));
        }
    }

}
