/*
 Copyright (c) 2010 Technische Universitt Darmstadt.
 
 Permission is hereby granted to use, copy, modify, and distribute this
 software, provided that this copyright notice remains intact.
 */

package ptolemy.batch;

import java.io.File;

import tcl.lang.Command;
import tcl.lang.Interp;
import tcl.lang.TclException;
import tcl.lang.TclNumArgsException;
import tcl.lang.TclObject;

import ptolemy.moml.MoMLParser;
import ptolemy.moml.filter.*;
import ptolemy.actor.CompositeActor;

/**
 * This class implements a new Tcl command "ptloadmodel"
 * that loads a Ptolemy II model from a file and returns a handle to that model.
 * 
 * @author cspies
 */
public class PtLoadModelCommand implements Command {

    /**
     * Invoked by the interpreter to handle the "ptloadmodel" command.
     * 
     * The command syntax is:
     * 
     *     ptloadmodel name
     * 
     * The MoML file specified by 'name' is opened and read.
     * If the file exists and contains a valid Ptolemy II model,
     * a handle to the top-level entity is returned.
     * If the file does not exist or any error occurs while processing it,
     * an error message is returned as the Tcl result object.
     * 
     * @param interp The interpreter used to execute the command.
     * @param objv The parameters of the model.
     *             Besides the command name, one parameter is expected.
     */
    public void cmdProc(Interp interp, TclObject[] objv) throws TclException {
        // local variables:
        MoMLParser parser;
        CompositeActor toplevel;
        String momlFileName;
        
        // get input:
        if (objv.length != 2) throw new TclNumArgsException(interp, 1, objv, "fileName"); // make sure there is exactly one command and one parameter
        momlFileName = objv[1].toString();
        
        // perform action:
        try {
            parser = new MoMLParser(); // instantiate new parser
            MoMLParser.setMoMLFilters(BackwardCompatibility.allFilters()); // set filters to handle backward compatibility
            MoMLParser.addMoMLFilter(new RemoveGraphicalClasses()); // disallow graphical classes because there is no GUI
            toplevel = (CompositeActor)parser.parse(null, new File(momlFileName).toURI().toURL());
        } catch (Exception ex) {
            throw new TclException(interp, String.format("Error reading '%s': %s", momlFileName, ex.toString()));
        }
        
        // produce output:
        interp.setResult(TclPtolemyActor.newInstance(toplevel)); // return a wrapper object containing the top-level entity
    }

}
