/*
 Copyright (c) 2010--2011 Technische Universitt Darmstadt.
 
 Permission is hereby granted to use, copy, modify, and distribute this
 software, provided that this copyright notice remains intact.
 */

package ptolemy.batch;

import java.util.Date;
import java.text.SimpleDateFormat;

import tcl.lang.Command;
import tcl.lang.Interp;
import tcl.lang.TclException;
import tcl.lang.TclNumArgsException;
import tcl.lang.TclObject;

import ptolemy.actor.CompositeActor;
import ptolemy.actor.ExecutionListener;
import ptolemy.actor.Manager;
import ptolemy.kernel.util.IllegalActionException;

/**
 * This class implements a new Tcl command "ptrunmodel"
 * that runs a Ptolemy II model and waits for the execution to end.
 * 
 * @author cspies
 */
public class PtRunModelCommand implements Command, ExecutionListener {
    
    /**
     * The name for the new manager if one needs to be created.
     */
    private final static String MANAGER_NAME = "manager";
    
    /**
     * The most recent exception causing the model to terminate execution abnormally.
     * This will be used to prepare an error message for the Tcl interpreter.
     */
    private Throwable _recentException;
    
    /**
     * Indicates whether the execution has come to an end
     * (either because the model ran to completion, or because an exception occurred).
     * 
     * TRUE if the most recent model execution has ended,
     * FALSE otherwise.
     */
    private boolean _isExecutionOver;

    /**
     * Invoked by the interpreter to handle the "ptrunmodel" command.
     * 
     * The command syntax is:
     * 
     *     ptrunmodel handle
     * 
     * The Ptolemy II model specified by 'handle' is run.
     * The command waits for the execution of the model to end
     * and terminates only after this has happened. 
     * It returns nothing.
     * If the handle is invalid or any error occurs during execution,
     * an error message is returned as the Tcl result object.
     * 
     * @param interp The interpreter used to execute the command.
     * @param objv The parameters of the model.
     *             Besides the command name, one parameter is expected.
     */
    public void cmdProc(Interp interp, TclObject[] objv) throws TclException {
        // local variables:
        CompositeActor toplevel;
        Manager manager;
        
        // get input:
        if (objv.length != 2) throw new TclNumArgsException(interp, 1, objv, "actorHandle"); // make sure there is exactly one command and one parameter
        toplevel = TclPtolemyActor.get(interp, objv[1]);
        
        // perform action:
        try {
            manager = toplevel.getManager();
            if (manager == null) {
                manager = new Manager(toplevel.workspace(), MANAGER_NAME);
                toplevel.setManager(manager);
            }
            manager.addExecutionListener(this);
        } catch (IllegalActionException iaex) {
            throw new TclException(interp, String.format("Error setting up model: %s", iaex.toString()));
        }
        _recentException = null;
        _isExecutionOver = false;
        _printSystemTime();
        manager.run(); // run() does not throw exceptions, as opposed to execute()
        while (!_isExecutionOver) Thread.yield(); // wait for execution to finish
        if (_recentException == null) {
            System.out.println("The model ran to completion.");
        } else {
            System.out.println("The model stopped prematurely due to an error:");
            System.out.println(_recentException.getMessage());
            _recentException.printStackTrace();
            
            Throwable cause = _recentException.getCause();
            while (cause != null) {
            	System.out.println("Because:");
            	System.out.println(cause.getMessage());
            	cause.printStackTrace();
            	cause = cause.getCause(); // check for nested causes recursively
            }
        }
        
        // don't produce output:
        interp.resetResult();
    }
    
    /** 
     * Invoked by the model manager to report an execution failure.
     *  
     * The command terminates after this method is called.
     *
     * @param manager The manager controlling the execution.
     * @param throwable The throwable to report.
     */
    public void executionError(Manager manager, Throwable throwable) {
        _recentException = throwable;
        _isExecutionOver = true;
    }

    /** 
     * Invoked by the model manager to report that the execution has finished normally.
     * 
     * The command terminates after this method is called.
     *
     * @param manager The manager controlling the execution.
     */
    public void executionFinished(Manager manager) {
        _recentException = null;
        _isExecutionOver = true;
    }

    /** 
     * Invoked by the model manager to report a change in its state.
     * 
     * This method does nothing.
     *
     * @param manager The manager controlling the execution.
     * @see Manager#getState()
     */
    public void managerStateChanged(Manager manager) {}
    
    /**
     * Print out the current system time!
     * 
     * This is used to highlight the start of model execution.
     */
    private static void _printSystemTime() {
        Date now = new Date();
        SimpleDateFormat format = new SimpleDateFormat("@ EEEE, MMMM d, yyyy --- HH:mm:ss zzz");
        System.out.println(format.format(now));
    }

}
