/*
 Copyright (c) 2010--2012 Technische Universitt Darmstadt.
 
 Permission is hereby granted to use, copy, modify, and distribute this
 software, provided that this copyright notice remains intact.
 */

package ptolemy.batch;

import tcl.lang.Command;
import tcl.lang.Interp;
import tcl.lang.TclException;
import tcl.lang.TclNumArgsException;
import tcl.lang.TclObject;

import ptolemy.actor.CompositeActor;
import ptolemy.data.expr.Parameter;
import ptolemy.kernel.util.IllegalActionException;

/**
 * This class implements a new Tcl command "ptsetparameter"
 * that manipulates a parameter within a Ptolemy II model.
 * 
 * @author cspies
 */
public class PtSetParameterCommand implements Command {

    /**
     * Invoked by the interpreter to handle the "ptsetparameter" command.
     * 
     * The command syntax is:
     * 
     *     ptsetparameter handle name value
     * 
     * The Ptolemy II model specified by 'handle'
     * is searched for the parameter specified by 'name',
     * and that parameter is set to 'value'.
     * The command returns nothing.
     * If the handle is invalid, the parameter does not exist or does not accept the given value,
     * an error message is returned as the Tcl result object.
     * 
     * @param interp The interpreter used to execute the command.
     * @param objv The parameters of the model.
     *             Besides the command name, three parameters are expected.
     */
    public void cmdProc(Interp interp, TclObject[] objv) throws TclException {
        // local variables:
        CompositeActor toplevel;
        Parameter parameter;
        String parameterName;
        String parameterValue;
        
        // get input:
        if (objv.length != 4) throw new TclNumArgsException(interp, 1, objv, "actorHandle parameterName parameterValue"); // make sure there is exactly one command and three parameters
        toplevel = TclPtolemyActor.get(interp, objv[1]);
        parameterName = objv[2].toString();
        parameterValue = objv[3].toString();
        
        // perform action:
        parameter = (Parameter)toplevel.getAttribute(parameterName);
        if (parameter == null) throw new TclException(interp, String.format("No parameter named '%s' exists inside actor '%s'!", parameterName, toplevel.getName()));
        parameter.setExpression(parameterValue);
        try {
        	parameterValue = parameter.getToken().toString();
        } catch (IllegalActionException iaex) {
        	throw new TclException(interp, String.format("Error accessing parameter '%s': %s", parameterName, iaex.toString()));
        }
        
        // produce output:
        System.out.println(String.format("Parameter '%s' has been set to %s.", parameterName, parameterValue));
        interp.resetResult();
    }

}
