/*
 Copyright (c) 2010 Technische Universitt Darmstadt.
 
 Permission is hereby granted to use, copy, modify, and distribute this
 software, provided that this copyright notice remains intact.
 */

package ptolemy.batch;

import ptolemy.actor.CompositeActor;

import tcl.lang.InternalRep;
import tcl.lang.Interp;
import tcl.lang.TclException;
import tcl.lang.TclObject;

/**
 * This class serves as a wrapper type for Ptolemy II "Composite Actors" in Tcl.
 *
 * @author Christopher Spies
 */
public class TclPtolemyActor implements InternalRep {
    
    private CompositeActor value;
    
    /**
     * Construct a TclPtolemyActor representation for the given actor!
     *
     * @param actor The actor.
     */
    private TclPtolemyActor(CompositeActor actor) {
        value = actor;
    }

    /**
     * Free the state associated with this object!
     */
    public void dispose() {
        this.value = null;
    }

    /**
     * Return a dupilcate of the current object!
     * 
     * @return Another TclPtolemyActor object representing the identical actor.
     */
    public InternalRep duplicate() {
        return new TclPtolemyActor(this.value);
    }
    
    /**
     * Retrieve the Ptolemy II actor from its Tcl representation!
     * 
     * @param interpreter The current interpreter.
     * @param tclobject The wrapper object to retrieve the actor from.
     * @return The actor contained in the wrapper object.
     */
    public static CompositeActor get(Interp interpreter, TclObject tclobject) throws TclException {
        InternalRep internalRepresentation;
        TclPtolemyActor wrapper;
        
        internalRepresentation = tclobject.getInternalRep();
        if (internalRepresentation instanceof TclPtolemyActor) {
            wrapper = (TclPtolemyActor)internalRepresentation;
            return wrapper.value;
        } else {
            // raise an exception if the given Tcl object does not contain a TclPtolemyActor internal representation
            throw new TclException(interpreter, "Handle to Ptolemy II actor expected");
        }
    }
    
    /**
     * Create a new instance of a TclObject
     * with an internal representation for the given actor! 
     * 
     * @param actor The actor.
     * @return A Tcl object representing a handle to that actor.
     */
    public static TclObject newInstance(CompositeActor actor) {
        return new TclObject(new TclPtolemyActor(actor));
    }
    
    /**
     * Return the string representation of this object!
     * 
     * @return The string representation of this object, as implemented by {@link CompositeActor#toString()}.
     */
    public String toString() {
        return this.value.toString();
    }

}
