/*
 Copyright (c) 2012 Technische Universitt Darmstadt.
 
 Derived work, based on the Integrator class by the University of California
 (see copyright notice below).
 */
/* The integrator in the CT domain.

 Copyright (c) 1998-2007 The Regents of the University of California.
 All rights reserved.
 Permission is hereby granted, without written agreement and without
 license or royalty fees, to use, copy, modify, and distribute this
 software and its documentation for any purpose, provided that the above
 copyright notice and the following two paragraphs appear in all copies
 of this software.

 IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
 THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

 THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
 PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
 CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
 ENHANCEMENTS, OR MODIFICATIONS.

 PT_COPYRIGHT_VERSION_2
 COPYRIGHTENDKEY

 */
package ptolemy.domains.ct.lib;

import ptolemy.data.DoubleToken;
import ptolemy.data.expr.Parameter;
import ptolemy.data.type.BaseType;
import ptolemy.domains.ct.kernel.CTBaseIntegrator;
import ptolemy.kernel.CompositeEntity;
import ptolemy.kernel.util.Attribute;
import ptolemy.kernel.util.IllegalActionException;
import ptolemy.kernel.util.NameDuplicationException;

//////////////////////////////////////////////////////////////////////////
//// AntiWindUpIntegrator

/**
 <p>
  An anti-wind-up integrator in the continuous time (CT) domain.
  This actor has one input port and one output port and behaves like a regular integrator,
  except that the value range for the output is limited and the output is saturated.
  This is useful e. g. in the modeling of integral control systems with anti-wind-up mechanisms in place.
 <p>
  Analytically, this behavior is identical to performing a LIMIT operation on the output.
  Numerically, however, problems may arise if the integrator state keeps increasing
  until it becomes much larger than typical state changes or than the value it is divided by,
  leading to a loss of precision. For this reason, the present actor was created.

 @author Christopher Spies
 @see ptolemy.domains.ct.lib.Integrator
 */
public class AntiWindUpIntegrator extends Integrator {
    
    /** Construct an integrator.
     * @see ptolemy.domains.ct.lib.Integrator
     * @param container The container.
     * @param name The name.
     * @exception NameDuplicationException If there is another entity with the same name.
     * @exception IllegalActionException If there was an internal problem.
     */
    public AntiWindUpIntegrator(CompositeEntity container, String name) throws NameDuplicationException, IllegalActionException {
        super(container, name);
        lowerBound = new Parameter(this, "lowerBound", new DoubleToken(Double.NEGATIVE_INFINITY));
        lowerBound.setTypeEquals(BaseType.DOUBLE);
        upperBound = new Parameter(this, "upperBound", new DoubleToken(Double.POSITIVE_INFINITY));
        upperBound.setTypeEquals(BaseType.DOUBLE);
    }
    
    ///////////////////////////////////////////////////////////////////
    ////                     ports and parameters                  ////
    
    /** The lower limit of the allowed output range. The default value is -INFINITY.
     *  The default range of -INFINITY to +INFINITY make this actor behave like a regular integrator.
     */
    public Parameter lowerBound;
    
    /** The upper limit of the allowed output range. The default value is +INFINITY.
     *  The default range of -INFINITY to +INFINITY make this actor behave like a regular integrator.
     */
    public Parameter upperBound;

    ///////////////////////////////////////////////////////////////////
    ////                         public methods                    ////
    
    /** If the specified attribute is <i>lowerLimit</i> or <i>upperLimit</i>,
     *  then change the corresponding internal variables for future use.
     *  
     *  @param attribute The attribute that has changed.
     *  @exception IllegalActionException If the new parameter value
     *  is not valid, especially if the upper bound is less than the lower bound
     */
    public void attributeChanged(Attribute attribute) throws IllegalActionException {
        double initialValue = ((DoubleToken)initialState.getToken()).doubleValue();
        
        if (attribute == lowerBound) {
            _minimumValue = ((DoubleToken)lowerBound.getToken()).doubleValue();
            if (_minimumValue >= _maximumValue) throw new IllegalActionException(this, INVALID_RANGE_MESSAGE);
            if (_minimumValue > initialValue) throw new IllegalActionException(this, INVALID_INITIAL_STATE_MESSAGE);
        } else if (attribute == upperBound) {
            _maximumValue = ((DoubleToken)upperBound.getToken()).doubleValue();
            if (_maximumValue <= _minimumValue) throw new IllegalActionException(this, INVALID_RANGE_MESSAGE);
            if (_maximumValue < initialValue) throw new IllegalActionException(this, INVALID_INITIAL_STATE_MESSAGE);
        } else {
            if (attribute == initialState) {
                if (initialValue < _minimumValue
                 || initialValue > _maximumValue)
                    throw new IllegalActionException(this, INVALID_INITIAL_STATE_MESSAGE);
            }
            super.attributeChanged(attribute);
        }
    }
    
    /** Initialize the integrator. Check for the existence of director and ODE
     *  solver. Update initial state parameter. Set the initial state to
     *  the tentative state and the state. Set tentative derivative to 0.0.
     *  Clear the history.
     *
     *  @exception If such an exception is thrown by the superclass.
     *  @see CTBaseIntegrator#initialize()
     */
    public void initialize() throws IllegalActionException {
        super.initialize();
        _minimumValue = ((DoubleToken)lowerBound.getToken()).doubleValue();
        _maximumValue = ((DoubleToken)upperBound.getToken()).doubleValue();
        setTentativeState(getTentativeState()); // make sure setTentativeState() is called to adjust the state
    }
    
    /** Set the tentative state. Tentative state is the state that
     *  the ODE solver resolved in one step. It may not
     *  be the final state due to error control or event detection.
     *  @param value The value to be set.
     *  @see #getTentativeState
     *  @see CTBaseIntegrator#setTentativeState(double)
     */
    public void setTentativeState(double value) {
        if (value < _minimumValue) value = _minimumValue;
        if (value > _maximumValue) value = _maximumValue;
        super.setTentativeState(value);
    }
    
    ///////////////////////////////////////////////////////////////////
    ////                         private variables                 ////
    
    /**
     * The lower boundary of the allowed output range.
     */
    private double _minimumValue;
    
    /**
     * The lower boundary of the allowed output range.
     */
    private double _maximumValue;
    
    /**
     * The error message to be shown when lowerBound and upperBound are in the wrong order.
     * @see #lowerBound
     * @see #upperBound
     */
    private final static String INVALID_RANGE_MESSAGE = "lowerBound must be less than upperBound";
    
    /**
     * The error message to be shown when the initial state is not within the range defined by lowerBound and upperBound.
     * @see CTBaseIntegrator#initialState
     * @see #lowerBound
     * @see #upperBound
     */
    private final static String INVALID_INITIAL_STATE_MESSAGE = "initialState must be greater than or equal to lowerBound and less than or equal to upperBound";

}
