/*
 Copyright (c) 2007 Technische Universitt Darmstadt.
 
 Derived work, based on the Queue class by the University of California
 (see copyright notice below).
 */
/* An actor that implements a queue of events.

 Copyright (c) 1998-2006 The Regents of the University of California.
 All rights reserved.
 Permission is hereby granted, without written agreement and without
 license or royalty fees, to use, copy, modify, and distribute this
 software and its documentation for any purpose, provided that the above
 copyright notice and the following two paragraphs appear in all copies
 of this software.

 IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
 THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

 THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
 PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
 CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
 ENHANCEMENTS, OR MODIFICATIONS.

 PT_COPYRIGHT_VERSION_2
 COPYRIGHTENDKEY

 */
package ptolemy.domains.de.lib;

import ptolemy.actor.TypedIOPort;
import ptolemy.actor.lib.Transformer;
import ptolemy.data.IntToken;
import ptolemy.data.Token;
import ptolemy.data.expr.Parameter;
import ptolemy.data.type.BaseType;
import ptolemy.kernel.CompositeEntity;
import ptolemy.kernel.util.Attribute;
import ptolemy.kernel.util.IllegalActionException;
import ptolemy.kernel.util.NameDuplicationException;
import ptolemy.kernel.util.StringAttribute;
import ptolemy.kernel.util.Workspace;

//////////////////////////////////////////////////////////////////////////
//// TappedCircularBuffer

/**
 This actor implements a tapped circular buffer.
 When a token is received on the <i>input</i> port,
 it is put into the buffer,
 and the oldest element is removed from the buffer.
 When the <i>taps</i> port receives a token on any channel,
 its value is used as an index into the circular buffer,
 and the element with that index is produced on the corresponding output channel.
 If there is no such element, then no output is produced.  
 In this circumstance, if <i>persistentTrigger</i> is true,
 then the next time an input is received, it is sent immediately to the output.
 
 <p>
 The inputs can be of any token type, and the output
 is constrained to be of a type at least that of the input.

 @author Christopher Spies
 */
public class TappedCircularBuffer extends Transformer {
    
    /** The default capacity. */
    private final static int DEFAULT_CAPACITY = 1;
    
    /** Construct an actor with the given container and name.
     *  @param container The container.
     *  @param name The name of this actor.
     *  @exception IllegalActionException If the actor cannot be contained
     *   by the proposed container.
     *  @exception NameDuplicationException If the container already has an
     *   actor with this name.
     */
    public TappedCircularBuffer(CompositeEntity container, String name)
            throws NameDuplicationException, IllegalActionException {
        super(container, name);
        output.setTypeAtLeast(input);
        output.setMultiport(true);

        taps = new TypedIOPort(this, "trigger", true, false);
        taps.setMultiport(true);
        taps.setTypeEquals(BaseType.INT); // taps are indices into the buffer, and must be integer
        
        _buffer = new Token[DEFAULT_CAPACITY]; // must be initialized before any parameters are instantiated
        _head = 0;
        
        StringAttribute cardinality = new StringAttribute(taps, "_cardinal");
        cardinality.setExpression("SOUTH"); // put the port to the bottom of the icon
        
        defaultValue = new Parameter(this, "defaultValue");
        defaultValue.setExpression("0.0");

        capacity = new Parameter(this, "capacity");
        capacity.setTypeEquals(BaseType.INT);
        capacity.setExpression(String.valueOf(DEFAULT_CAPACITY));
    }

    ///////////////////////////////////////////////////////////////////
    ////                     ports and parameters                  ////
    
    /** The default value to fill the buffer with initially. 
     */
    public Parameter defaultValue;

    /** The capacity of the queue. If the value is positive, then
     *  it specifies the capacity of the queue. If it is negative
     *  or 0, then it specifies that the capacity is infinite.
     *  This is an integer with default 0.
     */
    public Parameter capacity;

    /** The trigger port, which has undeclared type. If this port
     *  receives a token, then the oldest token in the queue
     *  will be emitted on the <i>output</i> port.
     */
    public TypedIOPort taps;

    ///////////////////////////////////////////////////////////////////
    ////                         public methods                    ////

    /** React to a change in an attribute.  If the attribute is
     *  <i>capacity</i>, then change the capacity of the queue.
     *  If the size of the queue currently exceeds the specified
     *  capacity, then throw an exception.
     *  @param attribute The attribute that changed.
     *  @exception IllegalActionException If the current size
     *   of the queue exceeds the specified capacity.
     */
    public void attributeChanged(Attribute attribute)
            throws IllegalActionException {
        if (attribute == defaultValue || attribute == capacity) {
            int newCapacity = ((IntToken) capacity.getToken()).intValue();
            if (newCapacity <= 0) {
                throw new IllegalActionException(this, "Buffer size must be positive!");
            } else {
                Token[] newBuffer = new Token[newCapacity];
                for (int index=0; index<_buffer.length && index<newCapacity; ++index) newBuffer[index] = _buffer[index];
                for (int index=_buffer.length; index<newCapacity; ++index) newBuffer[index] = defaultValue.getToken();
                _buffer = newBuffer;
                _head = _head % newCapacity;
            }
        } else {
            super.attributeChanged(attribute);
        }
    }

    /** Clone the actor into the specified workspace. This calls the
     *  base class and then sets the ports.
     *  @param workspace The workspace for the new object.
     *  @return A new actor.
     *  @exception CloneNotSupportedException If a derived class has
     *   has an attribute that cannot be cloned.
     */
    public Object clone(Workspace workspace) throws CloneNotSupportedException {
        TappedCircularBuffer newObject = (TappedCircularBuffer) super.clone(workspace);
        newObject._buffer = new Token[_buffer.length];
        for (int index=0; index<_buffer.length; ++index) newObject._buffer[index] = _buffer[index]; // arraycopy
        newObject._head = _head;
        newObject.output.setTypeAtLeast(newObject.input);
        return newObject;
    }

    /** Put a new input token into the buffer and/or produce output
     *  data from the buffer. Specifically, if there is a new token
     *  on the <i>input</i> port, then put it into the buffer.
     *  Then, if there are tokens in the <i>taps</i> port
     *  then send the corresponding elements to the <i>output</i> port.
     *  
     *  @exception IllegalActionException If getting tokens from input and
     *   trigger ports or sending token to output throws it.
     */
    public void fire() throws IllegalActionException {
        super.fire();

        if (input.hasToken(0)) {
            _head -= 1; if (_head < 0) _head += _buffer.length; // decrement head pointer to prepare for the next firing
            _buffer[_head] = input.get(0);
        }
        
        for (int channel=0; channel<taps.getWidth(); channel++) {
            if (taps.hasToken(channel)) {
                int index = ((IntToken)taps.get(channel)).intValue();
                index = (_head + index) % _buffer.length;
                output.send(channel, _buffer[index]);
            }
        }
    }

    /** Clear the buffer.
     *  @exception IllegalActionException If there is no director.
     */
    public void initialize() throws IllegalActionException {
        int size = ((IntToken)capacity.getToken()).intValue();
        _buffer = new Token[size];
        for (int index=0; index<size; ++index) _buffer[index] = defaultValue.getToken();
        _head = 0;
        super.initialize();
    }

    /** Return TRUE.
     *  @return True.
     *  @throws IllegalActionException If the superclass throws it.
     */
    public boolean postfire() throws IllegalActionException {
        return super.postfire();
    }

    /** Return TRUE as long as there are tokens on either the input or taps port.
     *  @exception IllegalActionException If there is no director.
     */
    public boolean prefire() throws IllegalActionException {
        if (input.getWidth() > 0) if (input.hasToken(0)) return true;
        for (int channel=0; channel<taps.getWidth(); ++channel) if (taps.hasToken(channel)) return true;
        return false;
    }

    ///////////////////////////////////////////////////////////////////
    ////                         protected variables               ////

    /** The circular buffer. */
    protected Token[] _buffer;
    
    /** The head pointer for the circular buffer. */
    protected int _head;

    ///////////////////////////////////////////////////////////////////
    ////                         private variables                 ////

}
